"""
Copyright (C) 2025 Interactive Brokers LLC. All rights reserved. This code is subject to the terms
 and conditions of the IB API Non-Commercial License or the IB API Commercial License, as applicable.
"""

import sys
import logging
import inspect
import time
import datetime

from decimal import Decimal

from ibapi.const import (
    UNSET_INTEGER,
    UNSET_DOUBLE,
    UNSET_LONG,
    UNSET_DECIMAL,
    DOUBLE_INFINITY,
    INFINITY_STR,
)

"""
Collection of misc tools
"""

logger = logging.getLogger(__name__)


# I use this just to visually emphasize it's a wrapper overridden method
def iswrapper(fn):
    return fn


class BadMessage(Exception):
    def __init__(self, text):
        self.text = text


class ClientException(Exception):
    def __init__(self, code, msg, text):
        self.code = code
        self.msg = msg
        self.text = text


class LogFunction(object):
    def __init__(self, text, logLevel):
        self.text = text
        self.logLevel = logLevel

    def __call__(self, fn):
        def newFn(origSelf, *args, **kwargs):
            if logger.isEnabledFor(self.logLevel):
                argNames = [
                    argName
                    for argName in inspect.getfullargspec(fn)[0]
                    if argName != "self"
                ]
                logger.log(
                    self.logLevel,
                    "{} {} {} kw:{}",
                    self.text,
                    fn.__name__,
                    [arg for arg in zip(argNames, args) if arg[1] is not origSelf],
                    kwargs,
                )
            fn(origSelf, *args)

        return newFn


def current_fn_name(parent_idx=0):
    # depth is 1 bc this is already a fn, so we need the caller
    return sys._getframe(1 + parent_idx).f_code.co_name


def setattr_log(self, var_name, var_value):
    # import code; code.interact(local=locals())
    logger.debug("%s %s %s=|%s|", self.__class__, id(self), var_name, var_value)
    super(self.__class__, self).__setattr__(var_name, var_value)


SHOW_UNSET = True


def decode(the_type, fields, show_unset=False, use_unicode=False):
    try:
        s = next(fields)
    except StopIteration:
        raise BadMessage("no more fields")

    logger.debug("decode %s %s", the_type, s)

    if the_type is Decimal:
        if (
            s is None
            or len(s) == 0
            or s.decode() == "2147483647"
            or s.decode() == "9223372036854775807"
            or s.decode() == "1.7976931348623157E308"
            or s.decode() == "-9223372036854775808"
        ):
            return UNSET_DECIMAL
        return the_type(s.decode())

    if the_type is str:
        if type(s) is str:
            return s
        if type(s) is bytes:
            return s.decode(
                "unicode-escape" if use_unicode else "UTF-8", errors="backslashreplace"
            )
        else:
            raise TypeError(
                "unsupported incoming type " + type(s) + " for desired type 'str"
            )

    orig_type = the_type
    if the_type is bool:
        the_type = int

    if the_type is float:
        if s.decode() == INFINITY_STR:
            return DOUBLE_INFINITY

    if show_unset:
        if s is None or len(s) == 0:
            if the_type is float:
                n = UNSET_DOUBLE
            elif the_type is int:
                n = UNSET_INTEGER
            else:
                raise TypeError("unsupported desired type for empty value" + the_type)
        else:
            n = the_type(s)
    else:
        n = the_type(s or 0)

    if orig_type is bool:
        n = n != 0

    return n


def ExerciseStaticMethods(klass):
    import types

    # import code; code.interact(local=dict(globals(), **locals()))
    for _, var in inspect.getmembers(klass):
        # print(name, var, type(var))
        if type(var) == types.FunctionType:
            print(f"Exercising: {var}:")
            print(var())
            print()

def isValidFloatValue(val: float) -> bool:
	return val != UNSET_DOUBLE

def isValidIntValue(val: int) -> bool:
    return val != UNSET_INTEGER

def isValidLongValue(val: int) -> bool:
    return val != UNSET_LONG

def isValidDecimalValue(val: Decimal) -> bool:
    return val != UNSET_DECIMAL

def floatMaxString(val: float):
    if val is None:
        return ""
    return (
        f"{val:.8f}".rstrip("0").rstrip(".").rstrip(",") if val != UNSET_DOUBLE else ""
    )


def longMaxString(val):
    return str(val) if val != UNSET_LONG else ""


def intMaxString(val):
    return str(val) if val != UNSET_INTEGER else ""


def isAsciiPrintable(val):
    return all(ord(c) >= 32 and ord(c) < 127 or ord(c) == 9 or ord(c) == 10 or ord(c) == 13 for c in val)


def decimalMaxString(val: Decimal):
    val = Decimal(val)
    return f"{val:f}" if val != UNSET_DECIMAL else ""


def isPegBenchOrder(orderType: str):
    return orderType in ("PEG BENCH", "PEGBENCH")


def isPegMidOrder(orderType: str):
    return orderType in ("PEG MID", "PEGMID")


def isPegBestOrder(orderType: str):
    return orderType in ("PEG BEST", "PEGBEST")


def log_(func, params, action):
    if logger.isEnabledFor(logging.INFO):
        if "self" in params:
            params = dict(params)
            del params["self"]
        logger.info(f"{action} {func} {params}")

def currentTimeMillis() :
    return round(time.time() * 1000)

def getTimeStrFromMillis(time: int):
    return datetime.datetime.fromtimestamp(time / 1000.0).strftime("%b %d, %Y %H:%M:%S.%f")[:-3] if time > 0 else ""

def listOfValues(cls):
    return list(map(lambda c: c, cls))

def getEnumTypeFromString(cls, stringIn):
    for item in cls:
        if item.value[0] == stringIn:
            return item
    return listOfValues(cls)[0]

def getEnumTypeName(cls, valueIn):
    for item in cls:
        if item == valueIn:
            return item.value[1]
    return listOfValues(cls)[0].value[1]




# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: ErrorMessage.proto
# Protobuf Python Version: 5.29.3
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    29,
    3,
    '',
    'ErrorMessage.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x12\x45rrorMessage.proto\x12\x08protobuf\"\xd8\x01\n\x0c\x45rrorMessage\x12\x0f\n\x02id\x18\x01 \x01(\x05H\x00\x88\x01\x01\x12\x16\n\terrorTime\x18\x02 \x01(\x03H\x01\x88\x01\x01\x12\x16\n\terrorCode\x18\x03 \x01(\x05H\x02\x88\x01\x01\x12\x15\n\x08\x65rrorMsg\x18\x04 \x01(\tH\x03\x88\x01\x01\x12$\n\x17\x61\x64vancedOrderRejectJson\x18\x05 \x01(\tH\x04\x88\x01\x01\x42\x05\n\x03_idB\x0c\n\n_errorTimeB\x0c\n\n_errorCodeB\x0b\n\t_errorMsgB\x1a\n\x18_advancedOrderRejectJsonB<\n\x16\x63om.ib.client.protobufB\x11\x45rrorMessageProto\xaa\x02\x0eIBApi.protobufb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'ErrorMessage_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\026com.ib.client.protobufB\021ErrorMessageProto\252\002\016IBApi.protobuf'
  _globals['_ERRORMESSAGE']._serialized_start=33
  _globals['_ERRORMESSAGE']._serialized_end=249
# @@protoc_insertion_point(module_scope)



"""
Copyright (C) 2019 Interactive Brokers LLC. All rights reserved. This code is subject to the terms
 and conditions of the IB API Non-Commercial License or the IB API Commercial License, as applicable.
"""

import unittest

from ibapi.enum_implem import Enum
from ibapi.utils import setattr_log


class UtilsTestCase(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_enum(self):
        e = Enum("ZERO", "ONE", "TWO")
        print(e.ZERO)
        print(e.toStr(e.ZERO))

    def test_setattr_log(self):
        class A:
            def __init__(self):
                self.n = 5

        A.__setattr__ = setattr_log
        a = A()
        print(a.n)
        a.n = 6
        print(a.n)

    def test_polymorphism(self):
        class A:
            def __init__(self):
                self.n = 5

            def m(self):
                self.n += 1

        class B(A):
            def m(self):
                self.n += 2

        o = B()
        print(o)
        # import code; code.interact(local=locals())


if "__main__" == __name__:
    unittest.main()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: Execution.proto
# Protobuf Python Version: 5.29.3
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    29,
    3,
    '',
    'Execution.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0f\x45xecution.proto\x12\x08protobuf\"\xaf\x06\n\tExecution\x12\x14\n\x07orderId\x18\x01 \x01(\x05H\x00\x88\x01\x01\x12\x13\n\x06\x65xecId\x18\x02 \x01(\tH\x01\x88\x01\x01\x12\x11\n\x04time\x18\x03 \x01(\tH\x02\x88\x01\x01\x12\x17\n\nacctNumber\x18\x04 \x01(\tH\x03\x88\x01\x01\x12\x15\n\x08\x65xchange\x18\x05 \x01(\tH\x04\x88\x01\x01\x12\x11\n\x04side\x18\x06 \x01(\tH\x05\x88\x01\x01\x12\x13\n\x06shares\x18\x07 \x01(\tH\x06\x88\x01\x01\x12\x12\n\x05price\x18\x08 \x01(\x01H\x07\x88\x01\x01\x12\x13\n\x06permId\x18\t \x01(\x03H\x08\x88\x01\x01\x12\x15\n\x08\x63lientId\x18\n \x01(\x05H\t\x88\x01\x01\x12\x1a\n\risLiquidation\x18\x0b \x01(\x08H\n\x88\x01\x01\x12\x13\n\x06\x63umQty\x18\x0c \x01(\tH\x0b\x88\x01\x01\x12\x15\n\x08\x61vgPrice\x18\r \x01(\x01H\x0c\x88\x01\x01\x12\x15\n\x08orderRef\x18\x0e \x01(\tH\r\x88\x01\x01\x12\x13\n\x06\x65vRule\x18\x0f \x01(\tH\x0e\x88\x01\x01\x12\x19\n\x0c\x65vMultiplier\x18\x10 \x01(\x01H\x0f\x88\x01\x01\x12\x16\n\tmodelCode\x18\x11 \x01(\tH\x10\x88\x01\x01\x12\x1a\n\rlastLiquidity\x18\x12 \x01(\x05H\x11\x88\x01\x01\x12#\n\x16isPriceRevisionPending\x18\x13 \x01(\x08H\x12\x88\x01\x01\x12\x16\n\tsubmitter\x18\x14 \x01(\tH\x13\x88\x01\x01\x12#\n\x16optExerciseOrLapseType\x18\x15 \x01(\x05H\x14\x88\x01\x01\x42\n\n\x08_orderIdB\t\n\x07_execIdB\x07\n\x05_timeB\r\n\x0b_acctNumberB\x0b\n\t_exchangeB\x07\n\x05_sideB\t\n\x07_sharesB\x08\n\x06_priceB\t\n\x07_permIdB\x0b\n\t_clientIdB\x10\n\x0e_isLiquidationB\t\n\x07_cumQtyB\x0b\n\t_avgPriceB\x0b\n\t_orderRefB\t\n\x07_evRuleB\x0f\n\r_evMultiplierB\x0c\n\n_modelCodeB\x10\n\x0e_lastLiquidityB\x19\n\x17_isPriceRevisionPendingB\x0c\n\n_submitterB\x19\n\x17_optExerciseOrLapseTypeB9\n\x16\x63om.ib.client.protobufB\x0e\x45xecutionProto\xaa\x02\x0eIBApi.protobufb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'Execution_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\026com.ib.client.protobufB\016ExecutionProto\252\002\016IBApi.protobuf'
  _globals['_EXECUTION']._serialized_start=30
  _globals['_EXECUTION']._serialized_end=845
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: OrderCondition.proto
# Protobuf Python Version: 5.29.3
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    29,
    3,
    '',
    'OrderCondition.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x14OrderCondition.proto\x12\x08protobuf\"\xea\x03\n\x0eOrderCondition\x12\x11\n\x04type\x18\x01 \x01(\x05H\x00\x88\x01\x01\x12$\n\x17isConjunctionConnection\x18\x02 \x01(\x08H\x01\x88\x01\x01\x12\x13\n\x06isMore\x18\x03 \x01(\x08H\x02\x88\x01\x01\x12\x12\n\x05\x63onId\x18\x04 \x01(\x05H\x03\x88\x01\x01\x12\x15\n\x08\x65xchange\x18\x05 \x01(\tH\x04\x88\x01\x01\x12\x13\n\x06symbol\x18\x06 \x01(\tH\x05\x88\x01\x01\x12\x14\n\x07secType\x18\x07 \x01(\tH\x06\x88\x01\x01\x12\x14\n\x07percent\x18\x08 \x01(\x05H\x07\x88\x01\x01\x12\x1a\n\rchangePercent\x18\t \x01(\x01H\x08\x88\x01\x01\x12\x12\n\x05price\x18\n \x01(\x01H\t\x88\x01\x01\x12\x1a\n\rtriggerMethod\x18\x0b \x01(\x05H\n\x88\x01\x01\x12\x11\n\x04time\x18\x0c \x01(\tH\x0b\x88\x01\x01\x12\x13\n\x06volume\x18\r \x01(\x05H\x0c\x88\x01\x01\x42\x07\n\x05_typeB\x1a\n\x18_isConjunctionConnectionB\t\n\x07_isMoreB\x08\n\x06_conIdB\x0b\n\t_exchangeB\t\n\x07_symbolB\n\n\x08_secTypeB\n\n\x08_percentB\x10\n\x0e_changePercentB\x08\n\x06_priceB\x10\n\x0e_triggerMethodB\x07\n\x05_timeB\t\n\x07_volumeB>\n\x16\x63om.ib.client.protobufB\x13OrderConditionProto\xaa\x02\x0eIBApi.protobufb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'OrderCondition_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\026com.ib.client.protobufB\023OrderConditionProto\252\002\016IBApi.protobuf'
  _globals['_ORDERCONDITION']._serialized_start=35
  _globals['_ORDERCONDITION']._serialized_end=525
# @@protoc_insertion_point(module_scope)



"""
Copyright (C) 2019 Interactive Brokers LLC. All rights reserved. This code is subject to the terms
 and conditions of the IB API Non-Commercial License or the IB API Commercial License, as applicable.
"""

import argparse
import datetime

from ibapi.client import EClient
from ibapi.common import *
from ibapi.contract import *
from ibapi.order import *
from ibapi.order_condition import *
from ibapi.order_state import *
from ibapi.scanner import ScannerSubscription
from ibapi.ticktype import TickType, TickTypeEnum
from ibapi.utils import *
from ibapi.wrapper import EWrapper


# import pdb; pdb.set_trace()
# import code; code.interact(local=locals())
# import code; code.interact(local=dict(globals(), **locals()))


class TestApp(EClient, EWrapper):
    def __init__(self):
        EClient.__init__(self, self)
        self.nextValidOrderId = None
        self.permId2ord = {}

    @iswrapper
    def nextValidId(self, orderId: int):
        super().nextValidId(orderId)
        logging.debug("setting nextValidOrderId: %d", orderId)
        self.nextValidOrderId = orderId

    def placeOneOrder(self):
        con = Contract()
        con.symbol = "AMD"
        con.secType = "STK"
        con.currency = "USD"
        con.exchange = "SMART"
        order = Order()
        order.action = "BUY"
        order.orderType = "LMT"
        order.tif = "GTC"
        order.totalQuantity = 3
        order.lmtPrice = 1.23
        self.placeOrder(self.nextOrderId(), con, order)

    def cancelOneOrder(self):
        pass

    def nextOrderId(self):
        id_ = self.nextValidOrderId
        self.nextValidOrderId += 1
        return id_

    @iswrapper
    def error(self, *args):
        super().error(*args)
        print(current_fn_name(), vars())

    @iswrapper
    def winError(self, *args):
        super().error(*args)
        print(current_fn_name(), vars())

    @iswrapper
    def openOrder(
        self, orderId: OrderId, contract: Contract, order: Order, orderState: OrderState
    ):
        super().openOrder(orderId, contract, order, orderState)
        print(current_fn_name(), vars())

        order.contract = contract
        self.permId2ord[order.permId] = order

    @iswrapper
    def openOrderEnd(self, *args):
        super().openOrderEnd()
        logging.debug("Received %d openOrders", len(self.permId2ord))

    @iswrapper
    def orderStatus(
        self,
        orderId: OrderId,
        status: str,
        filled: float,
        remaining: float,
        avgFillPrice: float,
        permId: int,
        parentId: int,
        lastFillPrice: float,
        clientId: int,
        whyHeld: str,
    ):
        super().orderStatus(
            orderId,
            status,
            Decimal(filled),
            Decimal(remaining),
            avgFillPrice,
            permId,
            parentId,
            lastFillPrice,
            clientId,
            whyHeld,
            0,
        )

    @iswrapper
    def tickPrice(self, tickerId: TickerId, tickType: TickType, price: float, attrib):
        super().tickPrice(tickerId, tickType, price, attrib)
        print(
            current_fn_name(),
            tickerId,
            TickTypeEnum.toStr(tickType),
            price,
            attrib,
            file=sys.stderr,
        )

    @iswrapper
    def tickSize(self, tickerId: TickerId, tickType: TickType, size: int):
        super().tickSize(tickerId, tickType, Decimal(size))
        print(
            current_fn_name(),
            tickerId,
            TickTypeEnum.toStr(tickType),
            size,
            file=sys.stderr,
        )

    @iswrapper
    def scannerParameters(self, xml: str):
        open("scanner.xml", "w").write(xml)


def main():
    cmdLineParser = argparse.ArgumentParser("api tests")
    # cmdLineParser.add_option("-c", action="store_true", dest="use_cache", default = False, help = "use the cache")
    # cmdLineParser.add_option("-f", action="store", type="string", dest="file", default="", help="the input file")
    cmdLineParser.add_argument(
        "-p",
        "--port",
        action="store",
        type=int,
        dest="port",
        default=4005,
        help="The TCP port to use",
    )
    args = cmdLineParser.parse_args()
    print("Using args", args)

    import logging

    logging.debug("Using args %s", args)
    # print(args)

    logging.debug("now is %s", datetime.datetime.now())
    logging.getLogger().setLevel(logging.ERROR)

    # enable logging when member vars are assigned
    import ibapi.utils as utils
    from ibapi.order import Order

    Order.__setattr__ = utils.setattr_log
    from ibapi.contract import Contract, DeltaNeutralContract

    Contract.__setattr__ = utils.setattr_log
    DeltaNeutralContract.__setattr__ = utils.setattr_log
    from ibapi.tag_value import TagValue

    TagValue.__setattr__ = utils.setattr_log
    TimeCondition.__setattr__ = utils.setattr_log
    ExecutionCondition.__setattr__ = utils.setattr_log
    MarginCondition.__setattr__ = utils.setattr_log
    PriceCondition.__setattr__ = utils.setattr_log
    PercentChangeCondition.__setattr__ = utils.setattr_log
    VolumeCondition.__setattr__ = utils.setattr_log

    # from inspect import signature as sig
    # import code; code.interact(local=dict(globals(), **locals()))
    # sys.exit(1)

    app = TestApp()
    app.connect("127.0.0.1", args.port, 0)

    app.reqCurrentTime()
    app.reqManagedAccts()
    app.reqAccountSummary(reqId=2, groupName="All", tags="NetLiquidation")

    app.reqAllOpenOrders()

    contract = Contract()
    contract.symbol = "AMD"
    contract.secType = "STK"
    contract.currency = "USD"
    contract.exchange = "SMART"
    # app.reqMarketDataType(1)
    # app.reqMktData(1001, contract, "", snapshot=True)
    # app.cancelMktData(1001)
    # app.reqExecutions(2001, ExecutionFilter())
    # app.reqContractDetails(3001, contract)
    # app.reqPositions()
    # app.reqIds(2)

    # app.reqMktDepth(4001, contract, 5, "")
    # app.cancelMktDepth(4001)

    # app.reqNewsBulletins(allMsgs=True)
    # app.cancelNewsBulletins()
    # app.requestFA(FaDataTypeEnum.GROUPS)

    # app.reqHistoricalData(5001, contract, "20161215 16:00:00", "2 D",
    #                             "1 hour", "TRADES", 0, 1, [])
    # app.cancelHistoricalData(5001)

    # app.reqFundamentalData(6001, contract, "ReportSnapshot")
    # app.cancelFundamentalData(6001)
    # app.queryDisplayGroups(7001)
    # app.subscribeToGroupEvents(7002, 1)
    # app.unsubscribeFromGroupEvents(7002)

    # app.reqScannerParameters()
    ss = ScannerSubscription()
    ss.instrument = "STK"
    ss.locationCode = "STK.US"
    ss.scanCode = "TOP_PERC_LOSE"
    # app.reqScannerSubscription(8001, ss, [])
    # app.cancelScannerSubscription(8001)
    # app.reqRealTimeBars(9001, contract, 5, "TRADES", 0, [])
    # app.cancelRealTimeBars(9001)
    # app.reqSecDefOptParams(10001, "AMD", "", "STK", 4391)
    # app.reqSoftDollarTiers(11001)
    # app.reqFamilyCodes()
    # app.reqMatchingSymbols(12001, "AMD")

    contract = Contract()
    contract.symbol = "AMD"
    contract.secType = "OPT"
    contract.exchange = "SMART"
    contract.currency = "USD"
    contract.lastTradeDateOrContractMonth = "20170120"
    contract.strike = 10
    contract.right = "C"
    contract.multiplier = "100"
    # Often, contracts will also require a trading class to rule out ambiguities
    contract.tradingClass = "AMD"
    # app.calculateImpliedVolatility(13001, contract, 1.3, 10.85)
    # app.calculateOptionPrice(13002, contract, 0.65, 10.85)

    app.run()


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: Contract.proto
# Protobuf Python Version: 5.29.3
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    29,
    3,
    '',
    'Contract.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


import ibapi.protobuf.ComboLeg_pb2 as ComboLeg__pb2
import ibapi.protobuf.DeltaNeutralContract_pb2 as DeltaNeutralContract__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0e\x43ontract.proto\x12\x08protobuf\x1a\x0e\x43omboLeg.proto\x1a\x1a\x44\x65ltaNeutralContract.proto\"\xdd\x06\n\x08\x43ontract\x12\x12\n\x05\x63onId\x18\x01 \x01(\x05H\x00\x88\x01\x01\x12\x13\n\x06symbol\x18\x02 \x01(\tH\x01\x88\x01\x01\x12\x14\n\x07secType\x18\x03 \x01(\tH\x02\x88\x01\x01\x12)\n\x1clastTradeDateOrContractMonth\x18\x04 \x01(\tH\x03\x88\x01\x01\x12\x13\n\x06strike\x18\x05 \x01(\x01H\x04\x88\x01\x01\x12\x12\n\x05right\x18\x06 \x01(\tH\x05\x88\x01\x01\x12\x17\n\nmultiplier\x18\x07 \x01(\x01H\x06\x88\x01\x01\x12\x15\n\x08\x65xchange\x18\x08 \x01(\tH\x07\x88\x01\x01\x12\x18\n\x0bprimaryExch\x18\t \x01(\tH\x08\x88\x01\x01\x12\x15\n\x08\x63urrency\x18\n \x01(\tH\t\x88\x01\x01\x12\x18\n\x0blocalSymbol\x18\x0b \x01(\tH\n\x88\x01\x01\x12\x19\n\x0ctradingClass\x18\x0c \x01(\tH\x0b\x88\x01\x01\x12\x16\n\tsecIdType\x18\r \x01(\tH\x0c\x88\x01\x01\x12\x12\n\x05secId\x18\x0e \x01(\tH\r\x88\x01\x01\x12\x18\n\x0b\x64\x65scription\x18\x0f \x01(\tH\x0e\x88\x01\x01\x12\x15\n\x08issuerId\x18\x10 \x01(\tH\x0f\x88\x01\x01\x12\x41\n\x14\x64\x65ltaNeutralContract\x18\x11 \x01(\x0b\x32\x1e.protobuf.DeltaNeutralContractH\x10\x88\x01\x01\x12\x1b\n\x0eincludeExpired\x18\x12 \x01(\x08H\x11\x88\x01\x01\x12\x1d\n\x10\x63omboLegsDescrip\x18\x13 \x01(\tH\x12\x88\x01\x01\x12%\n\tcomboLegs\x18\x14 \x03(\x0b\x32\x12.protobuf.ComboLegB\x08\n\x06_conIdB\t\n\x07_symbolB\n\n\x08_secTypeB\x1f\n\x1d_lastTradeDateOrContractMonthB\t\n\x07_strikeB\x08\n\x06_rightB\r\n\x0b_multiplierB\x0b\n\t_exchangeB\x0e\n\x0c_primaryExchB\x0b\n\t_currencyB\x0e\n\x0c_localSymbolB\x0f\n\r_tradingClassB\x0c\n\n_secIdTypeB\x08\n\x06_secIdB\x0e\n\x0c_descriptionB\x0b\n\t_issuerIdB\x17\n\x15_deltaNeutralContractB\x11\n\x0f_includeExpiredB\x13\n\x11_comboLegsDescripB8\n\x16\x63om.ib.client.protobufB\rContractProto\xaa\x02\x0eIBApi.protobufb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'Contract_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\026com.ib.client.protobufB\rContractProto\252\002\016IBApi.protobuf'
  _globals['_CONTRACT']._serialized_start=73
  _globals['_CONTRACT']._serialized_end=934
# @@protoc_insertion_point(module_scope)



"""
Copyright (C) 2019 Interactive Brokers LLC. All rights reserved. This code is subject to the terms
 and conditions of the IB API Non-Commercial License or the IB API Commercial License, as applicable.
"""

import unittest

from ibapi.order_condition import *


class ConditionOrderTestCase(unittest.TestCase):
    conds = [
        VolumeCondition(8314, "SMART", True, 1000000).And(),
        PercentChangeCondition(1111, "AMEX", True, 0.25).Or(),
        PriceCondition(
            PriceCondition.TriggerMethodEnum.DoubleLast, 2222, "NASDAQ", False, 4.75
        ).And(),
        TimeCondition(True, "20170101 09:30:00").And(),
        MarginCondition(False, 200000).Or(),
        ExecutionCondition("STK", "SMART", "AMD"),
    ]

    for cond in conds:
        print(cond, OrderCondition.__str__(cond))


if "__main__" == __name__:
    unittest.main()



"""
Copyright (C) 2025 Interactive Brokers LLC. All rights reserved. This code is subject to the terms
 and conditions of the IB API Non-Commercial License or the IB API Commercial License, as applicable.
"""

from ibapi.const import UNSET_DOUBLE
from ibapi.execution import ExecutionFilter

from ibapi.contract import Contract, DeltaNeutralContract, ComboLeg
from ibapi.order import Order
from ibapi.order_condition import OrderCondition, OperatorCondition, ContractCondition, PriceCondition, TimeCondition, MarginCondition, ExecutionCondition, VolumeCondition, PercentChangeCondition
from ibapi.tag_value import TagValue
from ibapi.utils import isValidIntValue, isValidFloatValue, isValidLongValue, isValidDecimalValue, decimalMaxString
from ibapi.order_condition import Create

from ibapi.protobuf.ComboLeg_pb2 import ComboLeg as ComboLegProto
from ibapi.protobuf.Contract_pb2 import Contract as ContractProto
from ibapi.protobuf.DeltaNeutralContract_pb2 import DeltaNeutralContract as DeltaNeutralContractProto
from ibapi.protobuf.OrderCancel_pb2 import OrderCancel as OrderCancelProto
from ibapi.protobuf.OrderCondition_pb2 import OrderCondition as OrderConditionProto
from ibapi.protobuf.PlaceOrderRequest_pb2 import PlaceOrderRequest as PlaceOrderRequestProto
from ibapi.protobuf.CancelOrderRequest_pb2 import CancelOrderRequest as CancelOrderRequestProto
from ibapi.protobuf.GlobalCancelRequest_pb2 import GlobalCancelRequest as GlobalCancelRequestProto
from ibapi.protobuf.ExecutionFilter_pb2 import ExecutionFilter as ExecutionFilterProto
from ibapi.protobuf.ExecutionRequest_pb2 import ExecutionRequest as ExecutionRequestProto
from ibapi.protobuf.Order_pb2 import Order as OrderProto
from ibapi.protobuf.SoftDollarTier_pb2 import SoftDollarTier as SoftDollarTierProto

@staticmethod
def createExecutionRequestProto(reqId: int, execFilter: ExecutionFilter) -> ExecutionRequestProto:
    executionFilterProto = ExecutionFilterProto()
    if isValidIntValue(execFilter.clientId): executionFilterProto.clientId = execFilter.clientId
    if execFilter.acctCode: executionFilterProto.acctCode = execFilter.acctCode
    if execFilter.time: executionFilterProto.time = execFilter.time
    if execFilter.symbol: executionFilterProto.symbol = execFilter.symbol
    if execFilter.secType: executionFilterProto.secType = execFilter.secType
    if execFilter.exchange: executionFilterProto.exchange = execFilter.exchange
    if execFilter.side: executionFilterProto.side = execFilter.side
    if isValidIntValue(execFilter.lastNDays): executionFilterProto.lastNDays = execFilter.lastNDays
    if execFilter.specificDates is not None and execFilter.specificDates: executionFilterProto.specificDates.extend(execFilter.specificDates)
    executionRequestProto = ExecutionRequestProto()
    if isValidIntValue(reqId): executionRequestProto.reqId = reqId
    executionRequestProto.executionFilter.CopyFrom(executionFilterProto)
    return executionRequestProto
 
@staticmethod
def createPlaceOrderRequestProto(orderId: int, contract: Contract, order: Order) -> PlaceOrderRequestProto:
    placeOrderRequestProto = PlaceOrderRequestProto()
    if isValidIntValue(orderId): placeOrderRequestProto.orderId = orderId
    contractProto = createContractProto(contract, order)
    if contractProto is not None: placeOrderRequestProto.contract.CopyFrom(contractProto)
    orderProto = createOrderProto(order)
    if orderProto is not None: placeOrderRequestProto.order.CopyFrom(orderProto)
    return placeOrderRequestProto

@staticmethod
def createContractProto(contract: Contract, order: Order) -> ContractProto:
    contractProto = ContractProto()
    if isValidIntValue(contract.conId): contractProto.conId = contract.conId
    if contract.symbol: contractProto.symbol = contract.symbol
    if contract.secType: contractProto.secType = contract.secType
    if contract.lastTradeDateOrContractMonth: contractProto.lastTradeDateOrContractMonth = contract.lastTradeDateOrContractMonth
    if isValidFloatValue(contract.strike): contractProto.strike = contract.strike
    if contract.right: contractProto.right = contract.right
    if contract.multiplier: contractProto.multiplier = float(contract.multiplier)
    if contract.exchange: contractProto.exchange = contract.exchange
    if contract.primaryExchange: contractProto.primaryExchange = contract.primaryExchange
    if contract.currency: contractProto.currency = contract.currency
    if contract.localSymbol: contractProto.localSymbol = contract.localSymbol
    if contract.tradingClass: contractProto.tradingClass = contract.tradingClass
    if contract.secIdType: contractProto.secIdType = contract.secIdType
    if contract.secId: contractProto.secId = contract.secId
    if contract.includeExpired: contractProto.includeExpired = contract.includeExpired
    if contract.comboLegsDescrip: contractProto.comboLegsDescrip = contract.comboLegsDescrip
    if contract.description: contractProto.description = contract.description
    if contract.issuerId: contractProto.issuerId = contract.issuerId

    comboLegProtoList = createComboLegProtoList(contract, order)
    if comboLegProtoList is not None and comboLegProtoList: contractProto.comboLegs.extend(comboLegProtoList)

    deltaNeutralContractProto = createDeltaNeutralContractProto(contract)
    if deltaNeutralContractProto is not None: contractProto.deltaNeutralContract.CopyFrom(deltaNeutralContractProto)

    return contractProto

@staticmethod
def createDeltaNeutralContractProto(contract: Contract) -> DeltaNeutralContractProto:
    deltaNeutralContractProto = None
    if contract.deltaNeutralContract is not None:
        deltaNeutralContract = contract.deltaNeutralContract
        deltaNeutralContractProto = DeltaNeutralContractProto()
        if isValidIntValue(deltaNeutralContract.conId): deltaNeutralContractProto.conId = deltaNeutralContract.conId
        if isValidFloatValue(deltaNeutralContract.delta): deltaNeutralContractProto.delta = deltaNeutralContract.delta
        if isValidFloatValue(deltaNeutralContract.price): deltaNeutralContractProto.price = deltaNeutralContract.price
    return deltaNeutralContractProto


@staticmethod
def createComboLegProtoList(contract: Contract, order: Order) -> list[ComboLegProto]:
    comboLegs = contract.comboLegs
    orderComboLegs = order.orderComboLegs
    comboLegProtoList = []
    if comboLegs is not None and comboLegs:
        for i, comboLeg in enumerate(comboLegs):
            perLegPrice = UNSET_DOUBLE
            if orderComboLegs is not None and i < len(orderComboLegs):
                perLegPrice = orderComboLegs[i].price
            comboLegProto = createComboLegProto(comboLeg, perLegPrice)
            comboLegProtoList.append(comboLegProto)
    return comboLegProtoList

@staticmethod
def createComboLegProto(comboLeg: ComboLeg, perLegPrice: float) -> ComboLegProto:
    comboLegProto = ComboLegProto()
    if isValidIntValue(comboLeg.conId): comboLegProto.conId = comboLeg.conId
    if isValidIntValue(comboLeg.ratio): comboLegProto.ratio = comboLeg.ratio
    if comboLeg.action: comboLegProto.action = comboLeg.action
    if comboLeg.exchange: comboLegProto.exchange = comboLeg.exchange
    if isValidIntValue(comboLeg.openClose): comboLegProto.openClose = comboLeg.openClose
    if isValidIntValue(comboLeg.shortSaleSlot): comboLegProto.shortSalesSlot = comboLeg.shortSaleSlot
    if comboLeg.designatedLocation: comboLegProto.designatedLocation = comboLeg.designatedLocation
    if isValidIntValue(comboLeg.exemptCode): comboLegProto.exemptCode = comboLeg.exemptCode
    if isValidFloatValue(perLegPrice): comboLegProto.perLegPrice = perLegPrice
    return comboLegProto

@staticmethod
def createOrderProto(order: Order) -> OrderProto:
    orderProto = OrderProto()
    if isValidIntValue(order.clientId): order.clientId = order.clientId
    if isValidLongValue(order.permId): orderProto.permId = order.permId
    if isValidIntValue(order.parentId): orderProto.parentId = order.parentId
    if order.action: orderProto.action = order.action
    if isValidDecimalValue(order.totalQuantity): orderProto.totalQuantity = decimalMaxString(order.totalQuantity)
    if isValidIntValue(order.displaySize): orderProto.displaySize = order.displaySize
    if order.orderType: orderProto.orderType = order.orderType
    if isValidFloatValue(order.lmtPrice): orderProto.lmtPrice = order.lmtPrice
    if isValidFloatValue(order.auxPrice): orderProto.auxPrice = order.auxPrice
    if order.tif: orderProto.tif = order.tif
    if order.account: orderProto.account = order.account
    if order.settlingFirm: orderProto.settlingFirm = order.settlingFirm
    if order.clearingAccount: orderProto.clearingAccount = order.clearingAccount
    if order.clearingIntent: orderProto.clearingIntent = order.clearingIntent
    if order.allOrNone: orderProto.allOrNone = order.allOrNone
    if order.blockOrder: orderProto.blockOrder = order.blockOrder
    if order.hidden: orderProto.hidden = order.hidden
    if order.outsideRth: orderProto.outsideRth = order.outsideRth
    if order.sweepToFill: orderProto.sweepToFill = order.sweepToFill
    if isValidFloatValue(order.percentOffset): orderProto.percentOffset = order.percentOffset
    if isValidFloatValue(order.trailingPercent): orderProto.trailingPercent = order.trailingPercent
    if isValidFloatValue(order.trailStopPrice): orderProto.trailStopPrice = order.trailStopPrice
    if isValidIntValue(order.minQty): orderProto.minQty = order.minQty
    if order.goodAfterTime: orderProto.goodAfterTime = order.goodAfterTime
    if order.goodTillDate: orderProto.goodTillDate = order.goodTillDate
    if order.ocaGroup: orderProto.ocaGroup = order.ocaGroup
    if order.orderRef: orderProto.orderRef = order.orderRef
    if order.rule80A: orderProto.rule80A = order.rule80A
    if isValidIntValue(order.ocaType): orderProto.ocaType = order.ocaType
    if isValidIntValue(order.triggerMethod): orderProto.triggerMethod = order.triggerMethod
    if order.activeStartTime: orderProto.activeStartTime = order.activeStartTime
    if order.activeStopTime: orderProto.activeStopTime = order.activeStopTime
    if order.faGroup: orderProto.faGroup = order.faGroup
    if order.faMethod: orderProto.faMethod = order.faMethod
    if order.faPercentage: orderProto.faPercentage = order.faPercentage
    if isValidFloatValue(order.volatility): orderProto.volatility = order.volatility
    if isValidIntValue(order.volatilityType): orderProto.volatilityType = order.volatilityType
    if order.continuousUpdate: orderProto.continuousUpdate = order.continuousUpdate
    if isValidIntValue(order.referencePriceType): orderProto.referencePriceType = order.referencePriceType
    if order.deltaNeutralOrderType: orderProto.deltaNeutralOrderType = order.deltaNeutralOrderType
    if isValidFloatValue(order.deltaNeutralAuxPrice): orderProto.deltaNeutralAuxPrice = order.deltaNeutralAuxPrice
    if isValidIntValue(order.deltaNeutralConId): orderProto.deltaNeutralConId = order.deltaNeutralConId
    if order.deltaNeutralOpenClose: orderProto.deltaNeutralOpenClose = order.deltaNeutralOpenClose
    if order.deltaNeutralShortSale: orderProto.deltaNeutralShortSale = order.deltaNeutralShortSale
    if isValidIntValue(order.deltaNeutralShortSaleSlot): orderProto.deltaNeutralShortSaleSlot = order.deltaNeutralShortSaleSlot
    if order.deltaNeutralDesignatedLocation: orderProto.deltaNeutralDesignatedLocation = order.deltaNeutralDesignatedLocation
    if isValidIntValue(order.scaleInitLevelSize): orderProto.scaleInitLevelSize = order.scaleInitLevelSize
    if isValidIntValue(order.scaleSubsLevelSize): orderProto.scaleSubsLevelSize = order.scaleSubsLevelSize
    if isValidFloatValue(order.scalePriceIncrement): orderProto.scalePriceIncrement = order.scalePriceIncrement
    if isValidFloatValue(order.scalePriceAdjustValue): orderProto.scalePriceAdjustValue = order.scalePriceAdjustValue
    if isValidIntValue(order.scalePriceAdjustInterval): orderProto.scalePriceAdjustInterval = order.scalePriceAdjustInterval
    if isValidFloatValue(order.scaleProfitOffset): orderProto.scaleProfitOffset = order.scaleProfitOffset
    if order.scaleAutoReset: orderProto.scaleAutoReset = order.scaleAutoReset
    if isValidIntValue(order.scaleInitPosition): orderProto.scaleInitPosition = order.scaleInitPosition
    if isValidIntValue(order.scaleInitFillQty): orderProto.scaleInitFillQty = order.scaleInitFillQty
    if order.scaleRandomPercent: orderProto.scaleRandomPercent = order.scaleRandomPercent
    if order.scaleTable: orderProto.scaleTable = order.scaleTable
    if order.hedgeType: orderProto.hedgeType = order.hedgeType
    if order.hedgeParam: orderProto.hedgeParam = order.hedgeParam

    if order.algoStrategy: orderProto.algoStrategy = order.algoStrategy
    fillTagValueList(order.algoParams, orderProto.algoParams)
    if order.algoId: orderProto.algoId = order.algoId

    fillTagValueList(order.smartComboRoutingParams, orderProto.smartComboRoutingParams)

    if order.whatIf: orderProto.whatIf = order.whatIf
    if order.transmit: orderProto.transmit = order.transmit
    if order.overridePercentageConstraints: orderProto.overridePercentageConstraints = order.overridePercentageConstraints
    if order.openClose: orderProto.openClose = order.openClose
    if isValidIntValue(order.origin): orderProto.origin = order.origin
    if isValidIntValue(order.shortSaleSlot): orderProto.shortSaleSlot = order.shortSaleSlot
    if order.designatedLocation: orderProto.designatedLocation = order.designatedLocation
    if isValidIntValue(order.exemptCode): orderProto.exemptCode = order.exemptCode
    if order.deltaNeutralSettlingFirm: orderProto.deltaNeutralSettlingFirm = order.deltaNeutralSettlingFirm
    if order.deltaNeutralClearingAccount: orderProto.deltaNeutralClearingAccount = order.deltaNeutralClearingAccount
    if order.deltaNeutralClearingIntent: orderProto.deltaNeutralClearingIntent = order.deltaNeutralClearingIntent
    if isValidIntValue(order.discretionaryAmt): orderProto.discretionaryAmt = order.discretionaryAmt
    if order.optOutSmartRouting: orderProto.optOutSmartRouting = order.optOutSmartRouting
    if isValidIntValue(order.exemptCode): orderProto.exemptCode = order.exemptCode
    if isValidFloatValue(order.startingPrice): orderProto.startingPrice = order.startingPrice
    if isValidFloatValue(order.stockRefPrice): orderProto.stockRefPrice = order.stockRefPrice
    if isValidFloatValue(order.delta): orderProto.delta = order.delta
    if isValidFloatValue(order.stockRangeLower): orderProto.stockRangeLower = order.stockRangeLower
    if isValidFloatValue(order.stockRangeUpper): orderProto.stockRangeUpper = order.stockRangeUpper
    if order.notHeld: orderProto.notHeld = order.notHeld

    fillTagValueList(order.orderMiscOptions, orderProto.orderMiscOptions)

    if order.solicited: orderProto.solicited = order.solicited
    if order.randomizeSize: orderProto.randomizeSize = order.randomizeSize
    if order.randomizePrice: orderProto.randomizePrice = order.randomizePrice
    if isValidIntValue(order.referenceContractId): orderProto.referenceContractId = order.referenceContractId
    if isValidFloatValue(order.peggedChangeAmount): orderProto.peggedChangeAmount = order.peggedChangeAmount
    if order.isPeggedChangeAmountDecrease: orderProto.isPeggedChangeAmountDecrease = order.isPeggedChangeAmountDecrease
    if isValidFloatValue(order.referenceChangeAmount): orderProto.referenceChangeAmount = order.referenceChangeAmount
    if order.referenceExchangeId: orderProto.referenceExchangeId = order.referenceExchangeId
    if order.adjustedOrderType: orderProto.adjustedOrderType = order.adjustedOrderType
    if isValidFloatValue(order.triggerPrice): orderProto.triggerPrice = order.triggerPrice
    if isValidFloatValue(order.adjustedStopPrice): orderProto.adjustedStopPrice = order.adjustedStopPrice
    if isValidFloatValue(order.adjustedStopLimitPrice): orderProto.adjustedStopLimitPrice = order.adjustedStopLimitPrice
    if isValidFloatValue(order.adjustedTrailingAmount): orderProto.adjustedTrailingAmount = order.adjustedTrailingAmount
    if isValidIntValue(order.adjustableTrailingUnit): orderProto.adjustableTrailingUnit = order.adjustableTrailingUnit
    if isValidFloatValue(order.lmtPriceOffset): orderProto.lmtPriceOffset = order.lmtPriceOffset

    orderConditionList = createConditionsProto(order)
    if orderConditionList is not None and orderConditionList: orderProto.conditions.extend(orderConditionList)
    if order.conditionsCancelOrder: orderProto.conditionsCancelOrder = order.conditionsCancelOrder
    if order.conditionsIgnoreRth: orderProto.conditionsIgnoreRth = order.conditionsIgnoreRth

    if order.modelCode: orderProto.modelCode = order.modelCode
    if order.extOperator: orderProto.extOperator = order.extOperator

    softDollarTier = createSoftDollarTierProto(order)
    if softDollarTier is not None: orderProto.softDollarTier.CopyFrom(softDollarTier)

    if isValidFloatValue(order.cashQty): orderProto.cashQty = order.cashQty
    if order.mifid2DecisionMaker: orderProto.mifid2DecisionMaker = order.mifid2DecisionMaker
    if order.mifid2DecisionAlgo: orderProto.mifid2DecisionAlgo = order.mifid2DecisionAlgo
    if order.mifid2ExecutionTrader: orderProto.mifid2ExecutionTrader = order.mifid2ExecutionTrader
    if order.mifid2ExecutionAlgo: orderProto.mifid2ExecutionAlgo = order.mifid2ExecutionAlgo
    if order.dontUseAutoPriceForHedge: orderProto.dontUseAutoPriceForHedge = order.dontUseAutoPriceForHedge
    if order.isOmsContainer: orderProto.isOmsContainer = order.isOmsContainer
    if order.discretionaryUpToLimitPrice: orderProto.discretionaryUpToLimitPrice = order.discretionaryUpToLimitPrice
    if order.usePriceMgmtAlgo is not None: orderProto.usePriceMgmtAlgo = 1 if order.usePriceMgmtAlgo else 0
    if isValidIntValue(order.duration): orderProto.duration = order.duration
    if isValidIntValue(order.postToAts): orderProto.postToAts = order.postToAts
    if order.advancedErrorOverride: orderProto.advancedErrorOverride = order.advancedErrorOverride
    if order.manualOrderTime: orderProto.manualOrderTime = order.manualOrderTime
    if isValidIntValue(order.minTradeQty): orderProto.minTradeQty = order.minTradeQty
    if isValidIntValue(order.minCompeteSize): orderProto.minCompeteSize = order.minCompeteSize
    if isValidFloatValue(order.competeAgainstBestOffset): orderProto.competeAgainstBestOffset = order.competeAgainstBestOffset
    if isValidFloatValue(order.midOffsetAtWhole): orderProto.midOffsetAtWhole = order.midOffsetAtWhole
    if isValidFloatValue(order.midOffsetAtHalf): orderProto.midOffsetAtHalf = order.midOffsetAtHalf
    if order.customerAccount: orderProto.customerAccount = order.customerAccount
    if order.professionalCustomer: orderProto.professionalCustomer = order.professionalCustomer
    if order.bondAccruedInterest: orderProto.bondAccruedInterest = order.bondAccruedInterest
    if order.includeOvernight: orderProto.includeOvernight = order.includeOvernight
    if isValidIntValue(order.manualOrderIndicator): orderProto.manualOrderIndicator = order.manualOrderIndicator
    if order.submitter: orderProto.submitter = order.submitter
    if order.autoCancelParent: orderProto.autoCancelParent = order.autoCancelParent
    if order.imbalanceOnly: orderProto.imbalanceOnly = order.imbalanceOnly

    return orderProto

def createConditionsProto(order: Order) -> list[OrderConditionProto]:
    orderConditionProtoList = []
    if order.conditions is not None and order.conditions:
        for orderCondition in order.conditions:
            conditionType = orderCondition.condType

            if OrderCondition.Price == conditionType:
                orderConditionProto = createPriceConditionProto(orderCondition)
            elif OrderCondition.Time == conditionType:
                orderConditionProto = createTimeConditionProto(orderCondition)
            elif OrderCondition.Margin == conditionType:
                orderConditionProto = createMarginConditionProto(orderCondition)
            elif OrderCondition.Execution == conditionType:
                orderConditionProto = createExecutionConditionProto(orderCondition)
            elif OrderCondition.Volume == conditionType:
                orderConditionProto = createVolumeConditionProto(orderCondition)
            elif OrderCondition.PercentChange == conditionType:
                orderConditionProto = createPercentChangeConditionProto(orderCondition)

            if orderConditionProto is not None: orderConditionProtoList.append(orderConditionProto)

    return orderConditionProtoList

@staticmethod
def createOrderConditionProto(orderCondition: OrderCondition) -> OrderConditionProto:
    conditionType = orderCondition.condType
    isConjunctionConnection = orderCondition.isConjunctionConnection
    orderConditionProto = OrderConditionProto()
    if isValidIntValue(conditionType): orderConditionProto.type = conditionType
    orderConditionProto.isConjunctionConnection = isConjunctionConnection
    return orderConditionProto

@staticmethod
def createOperatorConditionProto(operatorCondition: OperatorCondition) -> OrderConditionProto:
    orderConditionProto = createOrderConditionProto(operatorCondition)
    operatorConditionProto = OrderConditionProto()
    operatorConditionProto.MergeFrom(orderConditionProto)
    operatorConditionProto.isMore = operatorCondition.isMore
    return operatorConditionProto

@staticmethod
def createContractConditionProto(contractCondition: ContractCondition) -> OrderConditionProto:
    operatorConditionProto = createOperatorConditionProto(contractCondition)
    contractConditionProto = OrderConditionProto()
    contractConditionProto.MergeFrom(operatorConditionProto)
    if isValidIntValue(contractCondition.conId): contractConditionProto.conId = contractCondition.conId
    if contractCondition.exchange: contractConditionProto.exchange = contractCondition.exchange
    return contractConditionProto

@staticmethod
def createPriceConditionProto(priceCondition: PriceCondition) -> OrderConditionProto:
    contractConditionProto = createContractConditionProto(priceCondition)
    priceConditionProto = OrderConditionProto()
    priceConditionProto.MergeFrom(contractConditionProto)
    if isValidFloatValue(priceCondition.price): priceConditionProto.price = priceCondition.price
    if isValidIntValue(priceCondition.triggerMethod): priceConditionProto.triggerMethod = priceCondition.triggerMethod
    return priceConditionProto

@staticmethod
def createTimeConditionProto(timeCondition: TimeCondition) -> OrderConditionProto:
    operatorConditionProto = createOperatorConditionProto(timeCondition)
    timeConditionProto = OrderConditionProto()
    timeConditionProto.MergeFrom(operatorConditionProto)
    if timeCondition.time: timeConditionProto.time = timeCondition.time
    return timeConditionProto

@staticmethod
def createMarginConditionProto(marginCondition: MarginCondition) -> OrderConditionProto:
    operatorConditionProto = createOperatorConditionProto(marginCondition)
    marginConditionProto = OrderConditionProto()
    marginConditionProto.MergeFrom(operatorConditionProto)
    if isValidFloatValue(marginCondition.percent): marginConditionProto.percent = marginCondition.percent
    return marginConditionProto

@staticmethod
def createExecutionConditionProto(executionCondition: ExecutionCondition) -> OrderConditionProto:
    orderConditionProto = createOrderConditionProto(executionCondition)
    executionConditionProto = OrderConditionProto()
    executionConditionProto.MergeFrom(orderConditionProto)
    if executionCondition.secType: executionConditionProto.secType = executionCondition.secType
    if executionCondition.exchange: executionConditionProto.exchange = executionCondition.exchange
    if executionCondition.symbol: executionConditionProto.symbol = executionCondition.symbol
    return executionConditionProto

@staticmethod
def createVolumeConditionProto(volumeCondition: VolumeCondition) -> OrderConditionProto:
    contractConditionProto = createContractConditionProto(volumeCondition)
    volumeConditionProto = OrderConditionProto()
    volumeConditionProto.MergeFrom(contractConditionProto)
    if isValidIntValue(volumeCondition.volume): volumeConditionProto.volume = volumeCondition.volume
    return volumeConditionProto

@staticmethod
def createPercentChangeConditionProto(percentChangeCondition: PercentChangeCondition) -> OrderConditionProto:
    contractConditionProto = createContractConditionProto(percentChangeCondition)
    percentChangeConditionProto = OrderConditionProto()
    percentChangeConditionProto.MergeFrom(contractConditionProto)
    if isValidFloatValue(percentChangeCondition.changePercent): percentChangeConditionProto.changePercent = percentChangeCondition.changePercent
    return percentChangeConditionProto

@staticmethod
def createSoftDollarTierProto(order: Order) -> SoftDollarTierProto:
    softDollarTierProto = None
    tier = order.softDollarTier
    if tier is not None:
        softDollarTierProto = SoftDollarTierProto()
        if tier.name: softDollarTierProto.name = tier.name
        if tier.val: softDollarTierProto.value = tier.val
        if tier.displayName: softDollarTierProto.displayName = tier.displayName
    return softDollarTierProto

@staticmethod
def fillTagValueList(tagValueList: list, orderProtoMap: dict):
    if tagValueList is not None and tagValueList:
        for tagValue in tagValueList: 
            orderProtoMap[tagValue.tag] = tagValue.value

@staticmethod
def createCancelOrderRequestProto(orderId: int, orderCancel: OrderCancelProto) -> CancelOrderRequestProto:
    cancelOrderRequestProto = CancelOrderRequestProto()
    if isValidIntValue(orderId): cancelOrderRequestProto.orderId = orderId
    orderCancelProto = createOrderCancelProto(orderCancel)
    if orderCancelProto is not None: cancelOrderRequestProto.orderCancel.CopyFrom(orderCancelProto)
    return cancelOrderRequestProto

@staticmethod
def createGlobalCancelRequestProto(orderCancel: OrderCancelProto) -> GlobalCancelRequestProto:
    globalCancelRequestProto = GlobalCancelRequestProto()
    orderCancelProto = createOrderCancelProto(orderCancel)
    if orderCancelProto is not None: globalCancelRequestProto.orderCancel.CopyFrom(orderCancelProto)
    return globalCancelRequestProto

@staticmethod
def createOrderCancelProto(orderCancel: OrderCancelProto) -> OrderCancelProto:
    orderCancelProto = OrderCancelProto()
    if orderCancel.manualOrderCancelTime: orderCancelProto.manualOrderCancelTime = orderCancel.manualOrderCancelTime
    if orderCancel.extOperator: orderCancelProto.extOperator = orderCancel.extOperator
    if isValidIntValue(orderCancel.manualOrderIndicator): orderCancelProto.manualOrderIndicator = orderCancel.manualOrderIndicator
    return orderCancelProto



"""
Copyright (C) 2025 Interactive Brokers LLC. All rights reserved. This code is subject to the terms
 and conditions of the IB API Non-Commercial License or the IB API Commercial License, as applicable.
"""
from decimal import Decimal

from ibapi.contract import ComboLeg, Contract, DeltaNeutralContract
from ibapi.execution import Execution, OptionExerciseType
from ibapi.order import Order, OrderComboLeg
from ibapi.order_condition import OrderCondition, OperatorCondition, ContractCondition, PriceCondition, TimeCondition, MarginCondition, ExecutionCondition, VolumeCondition, PercentChangeCondition
from ibapi.order_state import OrderState, OrderAllocation
from ibapi.softdollartier import SoftDollarTier
from ibapi.tag_value import TagValue
from ibapi.utils import floatMaxString, getEnumTypeFromString, decimalMaxString

#from ibapi.protobuf.ComboLeg_pb2 import ComboLeg as ComboLegProto
from ibapi.protobuf.Contract_pb2 import Contract as ContractProto
from ibapi.protobuf.DeltaNeutralContract_pb2 import DeltaNeutralContract as DeltaNeutralContractProto
from ibapi.protobuf.Execution_pb2 import Execution as ExecutionProto
from ibapi.protobuf.Order_pb2 import Order as OrderProto
from ibapi.protobuf.OrderCondition_pb2 import OrderCondition as OrderConditionProto
from ibapi.protobuf.OrderState_pb2 import OrderState as OrderStateProto
from ibapi.protobuf.OrderAllocation_pb2 import OrderAllocation as OrderAllocatiionProto
from ibapi.protobuf.SoftDollarTier_pb2 import SoftDollarTier as SoftDollarTierProto

@staticmethod
def decodeContract(contractProto: ContractProto) -> Contract:
    contract = Contract()
    if contractProto.HasField('conId'): contract.conId = contractProto.conId
    if contractProto.HasField('symbol'): contract.symbol = contractProto.symbol
    if contractProto.HasField('secType'): contract.secType = contractProto.secType
    if contractProto.HasField('lastTradeDateOrContractMonth'): contract.lastTradeDateOrContractMonth = contractProto.lastTradeDateOrContractMonth
    if contractProto.HasField('strike'): contract.strike = contractProto.strike
    if contractProto.HasField('right'): contract.right = contractProto.right
    if contractProto.HasField('multiplier'): contract.multiplier = floatMaxString(contractProto.multiplier)
    if contractProto.HasField('exchange'): contract.exchange = contractProto.exchange
    if contractProto.HasField('currency'): contract.currency = contractProto.currency
    if contractProto.HasField('localSymbol'): contract.localSymbol = contractProto.localSymbol
    if contractProto.HasField('tradingClass'): contract.tradingClass = contractProto.tradingClass
    if contractProto.HasField('comboLegsDescrip'): contract.comboLegsDescrip = contractProto.comboLegsDescrip

    comboLegs = decodeComboLegs(contractProto)
    if comboLegs is not None and comboLegs: contract.comboLegs = comboLegs

    deltaNeutralContract = decodeDeltaNeutralContract(contractProto)
    if deltaNeutralContract is not None: contract.deltaNeutralContract = deltaNeutralContract

    return contract

@staticmethod
def decodeComboLegs(contractProto: ContractProto) -> list[ComboLeg]:
    comboLegs = []
    comboLegProtoList = contractProto.comboLegs
    if comboLegProtoList:
        for comboLegProto in comboLegProtoList:
            comboLeg = ComboLeg()
            if comboLegProto.HasField('conId'): comboLeg.conId = comboLegProto.conId
            if comboLegProto.HasField('ratio'): comboLeg.ratio = comboLegProto.ratio
            if comboLegProto.HasField('action'): comboLeg.action = comboLegProto.action
            if comboLegProto.HasField('exchange'): comboLeg.exchange = comboLegProto.exchange
            if comboLegProto.HasField('openClose'): comboLeg.openClose = comboLegProto.openClose
            if comboLegProto.HasField('shortSalesSlot'): comboLeg.shortSalesSlot = comboLegProto.shortSalesSlot
            if comboLegProto.HasField('designatedLocation'): comboLeg.designatedLocation = comboLegProto.designatedLocation
            if comboLegProto.HasField('exemptCode'): comboLeg.exemptCode = comboLegProto.exemptCode
            comboLegs.append(comboLeg)

    return comboLegs

@staticmethod
def decodeOrderComboLegs(contractProto: ContractProto) -> list[OrderComboLeg]:
    orderComboLegs = []
    comboLegProtoList = contractProto.comboLegs
    if comboLegProtoList:
        for comboLegProto in comboLegProtoList:
            orderComboLeg = OrderComboLeg()
            if comboLegProto.HasField('perLegPrice'): orderComboLeg.price = comboLegProto.perLegPrice
            orderComboLegs.append(orderComboLeg)

    return orderComboLegs

@staticmethod
def decodeDeltaNeutralContract(contractProto: ContractProto) -> DeltaNeutralContract:
    deltaNeutralContract = None
    deltaNeutralContractProto = DeltaNeutralContractProto()
    if contractProto.HasField('deltaNeutralContract'): deltaNeutralContractProto.CopyFrom(contractProto.deltaNeutralContract)
    if deltaNeutralContractProto is not None: 
        deltaNeutralContract = DeltaNeutralContract()
        if (deltaNeutralContractProto.HasField('conId')): deltaNeutralContract.conId = deltaNeutralContractProto.conId
        if (deltaNeutralContractProto.HasField('delta')): deltaNeutralContract.delta = deltaNeutralContractProto.delta
        if (deltaNeutralContractProto.HasField('price')): deltaNeutralContract.price = deltaNeutralContractProto.price

    return deltaNeutralContract

@staticmethod
def decodeExecution(executionProto: ExecutionProto) -> Execution:
    execution = Execution()
    if executionProto.HasField('orderId'): execution.orderId = executionProto.orderId
    if executionProto.HasField('clientId'): execution.clientId = executionProto.clientId
    if executionProto.HasField('execId'): execution.execId = executionProto.execId
    if executionProto.HasField('time'): execution.time = executionProto.time
    if executionProto.HasField('acctNumber'): execution.acctNumber = executionProto.acctNumber
    if executionProto.HasField('exchange'): execution.exchange = executionProto.exchange
    if executionProto.HasField('side'): execution.side = executionProto.side
    if executionProto.HasField('shares'): execution.shares = Decimal(executionProto.shares)
    if executionProto.HasField('price'): execution.price = executionProto.price
    if executionProto.HasField('permId'): execution.permId = executionProto.permId
    if executionProto.HasField('isLiquidation'): execution.liquidation = 1 if executionProto.isLiquidation else 0
    if executionProto.HasField('cumQty'): execution.cumQty = Decimal(executionProto.cumQty)
    if executionProto.HasField('avgPrice'): execution.avgPrice = executionProto.avgPrice
    if executionProto.HasField('orderRef'): execution.orderRef = executionProto.orderRef
    if executionProto.HasField('evRule'): execution.evRule = executionProto.evRule
    if executionProto.HasField('evMultiplier'): execution.evMultiplier = executionProto.evMultiplier
    if executionProto.HasField('modelCode'): execution.modelCode = executionProto.modelCode
    if executionProto.HasField('lastLiquidity'): execution.lastLiquidity = executionProto.lastLiquidity
    if executionProto.HasField('isPriceRevisionPending'): execution.pendingPriceRevision = executionProto.isPriceRevisionPending
    if executionProto.HasField('submitter'): execution.submitter = executionProto.submitter
    if executionProto.HasField('optExerciseOrLapseType'): execution.optExerciseOrLapseType = getEnumTypeFromString(OptionExerciseType, executionProto.optExerciseOrLapseType)
    return execution

@staticmethod
def decodeOrder(contractProto: ContractProto, orderProto: OrderProto) -> Order:
    order = Order()
    if orderProto.HasField('action'): order.action = orderProto.action
    if orderProto.HasField('totalQuantity'): order.totalQuantity = orderProto.totalQuantity
    if orderProto.HasField('orderType'): order.orderType = orderProto.orderType
    if orderProto.HasField('lmtPrice'): order.lmtPrice = orderProto.lmtPrice
    if orderProto.HasField('auxPrice'): order.auxPrice = orderProto.auxPrice
    if orderProto.HasField('tif'): order.tif = orderProto.tif
    if orderProto.HasField('ocaGroup'): order.ocaGroup = orderProto.ocaGroup
    if orderProto.HasField('account'): order.account = orderProto.account
    if orderProto.HasField('openClose'): order.openClose = orderProto.openClose
    if orderProto.HasField('origin'): order.origin = orderProto.origin
    if orderProto.HasField('orderRef'): order.orderRef = orderProto.orderRef
    if orderProto.HasField('clientId'): order.clientId = orderProto.clientId
    if orderProto.HasField('permId'): order.permId = orderProto.permId
    if orderProto.HasField('outsideRth'): order.outsideRth = orderProto.outsideRth
    if orderProto.HasField('hidden'): order.hidden = orderProto.hidden
    if orderProto.HasField('discretionaryAmt'): order.discretionaryAmt = orderProto.discretionaryAmt
    if orderProto.HasField('goodAfterTime'): order.goodAfterTime = orderProto.goodAfterTime
    if orderProto.HasField('faGroup'): order.faGroup = orderProto.faGroup
    if orderProto.HasField('faMethod'): order.faMethod = orderProto.faMethod
    if orderProto.HasField('faPercentage'): order.faPercentage = orderProto.faPercentage
    if orderProto.HasField('modelCode'): order.modelCode = orderProto.modelCode
    if orderProto.HasField('goodTillDate'): order.goodTillDate = orderProto.goodTillDate
    if orderProto.HasField('rule80A'): order.rule80A = orderProto.rule80A
    if orderProto.HasField('percentOffset'): order.percentOffset = orderProto.percentOffset
    if orderProto.HasField('settlingFirm'): order.settlingFirm = orderProto.settlingFirm
    if orderProto.HasField('shortSaleSlot'): order.shortSaleSlot = orderProto.shortSaleSlot
    if orderProto.HasField('designatedLocation'): order.designatedLocation = orderProto.designatedLocation
    if orderProto.HasField('exemptCode'): order.exemptCode = orderProto.exemptCode
    if orderProto.HasField('startingPrice'): order.startingPrice = orderProto.startingPrice
    if orderProto.HasField('stockRefPrice'): order.stockRefPrice = orderProto.stockRefPrice
    if orderProto.HasField('delta'): order.delta = orderProto.delta
    if orderProto.HasField('stockRangeLower'): order.stockRangeLower = orderProto.stockRangeLower
    if orderProto.HasField('stockRangeUpper'): order.stockRangeUpper = orderProto.stockRangeUpper
    if orderProto.HasField('displaySize'): order.displaySize = orderProto.displaySize
    if orderProto.HasField('blockOrder'): order.blockOrder = orderProto.blockOrder
    if orderProto.HasField('sweepToFill'): order.sweepToFill = orderProto.sweepToFill
    if orderProto.HasField('allOrNone'): order.allOrNone = orderProto.allOrNone
    if orderProto.HasField('minQty'): order.minQty = orderProto.minQty
    if orderProto.HasField('ocaType'): order.ocaType = orderProto.ocaType
    if orderProto.HasField('parentId'): order.parentId = orderProto.parentId
    if orderProto.HasField('triggerMethod'): order.triggerMethod = orderProto.triggerMethod
    if orderProto.HasField('volatility'): order.volatility = orderProto.volatility
    if orderProto.HasField('volatilityType'): order.volatilityType = orderProto.volatilityType
    if orderProto.HasField('deltaNeutralOrderType'): order.deltaNeutralOrderType = orderProto.deltaNeutralOrderType
    if orderProto.HasField('deltaNeutralAuxPrice'): order.deltaNeutralAuxPrice = orderProto.deltaNeutralAuxPrice
    if orderProto.HasField('deltaNeutralConId'): order.deltaNeutralConId = orderProto.deltaNeutralConId
    if orderProto.HasField('deltaNeutralSettlingFirm'): order.ddeltaNeutralSettlingFirm = orderProto.deltaNeutralSettlingFirm
    if orderProto.HasField('deltaNeutralClearingAccount'): order.deltaNeutralClearingAccount = orderProto.deltaNeutralClearingAccount
    if orderProto.HasField('deltaNeutralClearingIntent'): order.deltaNeutralClearingIntent = orderProto.deltaNeutralClearingIntent
    if orderProto.HasField('deltaNeutralOpenClose'): order.deltaNeutralOpenClose = orderProto.deltaNeutralOpenClose
    if orderProto.HasField('deltaNeutralShortSale'): order.deltaNeutralShortSale = orderProto.deltaNeutralShortSale
    if orderProto.HasField('deltaNeutralShortSaleSlot'): order.deltaNeutralShortSaleSlot = orderProto.deltaNeutralShortSaleSlot
    if orderProto.HasField('deltaNeutralDesignatedLocation'): order.deltaNeutralDesignatedLocation = orderProto.deltaNeutralDesignatedLocation
    if orderProto.HasField('continuousUpdate'): order.continuousUpdate = orderProto.continuousUpdate
    if orderProto.HasField('referencePriceType'): order.referencePriceType = orderProto.referencePriceType
    if orderProto.HasField('trailStopPrice'): order.trailStopPrice = orderProto.trailStopPrice
    if orderProto.HasField('trailingPercent'): order.trailingPercent = orderProto.trailingPercent

    orderComboLegs = decodeOrderComboLegs(contractProto)
    if orderComboLegs is not None and orderComboLegs: order.orderComboLegs = orderComboLegs

    order.smartComboRoutingParams = decodeTagValueList(orderProto.smartComboRoutingParams)

    if orderProto.HasField('scaleInitLevelSize'): order.scaleInitLevelSize = orderProto.scaleInitLevelSize
    if orderProto.HasField('scaleSubsLevelSize'): order.scaleSubsLevelSize = orderProto.scaleSubsLevelSize
    if orderProto.HasField('scalePriceIncrement'): order.scalePriceIncrement = orderProto.scalePriceIncrement
    if orderProto.HasField('scalePriceAdjustValue'): order.scalePriceAdjustValue = orderProto.scalePriceAdjustValue
    if orderProto.HasField('scalePriceAdjustInterval'): order.scalePriceAdjustInterval = orderProto.scalePriceAdjustInterval
    if orderProto.HasField('scaleProfitOffset'): order.scaleProfitOffset = orderProto.scaleProfitOffset
    if orderProto.HasField('scaleAutoReset'): order.scaleAutoReset = orderProto.scaleAutoReset
    if orderProto.HasField('scaleInitPosition'): order.scaleInitPosition = orderProto.scaleInitPosition
    if orderProto.HasField('scaleInitFillQty'): order.scaleInitFillQty = orderProto.scaleInitFillQty
    if orderProto.HasField('scaleRandomPercent'): order.scaleRandomPercent = orderProto.scaleRandomPercent
    if orderProto.HasField('hedgeType'): order.hedgeType = orderProto.hedgeType
    if orderProto.HasField('hedgeType') and orderProto.HasField('hedgeParam') and orderProto.hedgeType: order.hedgeParam = orderProto.hedgeParam
    if orderProto.HasField('optOutSmartRouting'): order.optOutSmartRouting = orderProto.optOutSmartRouting
    if orderProto.HasField('clearingAccount'): order.clearingAccount = orderProto.clearingAccount
    if orderProto.HasField('clearingIntent'): order.clearingIntent = orderProto.clearingIntent
    if orderProto.HasField('notHeld'): order.notHeld = orderProto.notHeld

    if orderProto.HasField('algoStrategy'): 
        order.algoStrategy = orderProto.algoStrategy
        order.algoParams = decodeTagValueList(orderProto.algoParams)

    if orderProto.HasField('solicited'): order.solicited = orderProto.solicited
    if orderProto.HasField('whatIf'): order.whatIf = orderProto.whatIf
    if orderProto.HasField('randomizeSize'): order.randomizeSize = orderProto.randomizeSize
    if orderProto.HasField('randomizePrice'): order.randomizePrice = orderProto.randomizePrice
    if orderProto.HasField('referenceContractId'): order.referenceContractId = orderProto.referenceContractId
    if orderProto.HasField('isPeggedChangeAmountDecrease'): order.isPeggedChangeAmountDecrease = orderProto.isPeggedChangeAmountDecrease
    if orderProto.HasField('peggedChangeAmount'): order.peggedChangeAmount = orderProto.peggedChangeAmount
    if orderProto.HasField('referenceChangeAmount'): order.referenceChangeAmount = orderProto.referenceChangeAmount
    if orderProto.HasField('referenceExchangeId'): order.referenceExchangeId = orderProto.referenceExchangeId

    conditions = decodeConditions(orderProto)
    if conditions is not None and conditions: order.conditions = conditions
    if orderProto.HasField('conditionsIgnoreRth'): order.conditionsIgnoreRth = orderProto.conditionsIgnoreRth
    if orderProto.HasField('conditionsCancelOrder'): order.conditionsCancelOrder = orderProto.conditionsCancelOrder

    if orderProto.HasField('adjustedOrderType'): order.adjustedOrderType = orderProto.adjustedOrderType
    if orderProto.HasField('triggerPrice'): order.triggerPrice = orderProto.triggerPrice
    if orderProto.HasField('lmtPriceOffset'): order.lmtPriceOffset = orderProto.lmtPriceOffset
    if orderProto.HasField('adjustedStopPrice'): order.adjustedStopPrice = orderProto.adjustedStopPrice
    if orderProto.HasField('adjustedStopLimitPrice'): order.adjustedStopLimitPrice = orderProto.adjustedStopLimitPrice
    if orderProto.HasField('adjustedTrailingAmount'): order.adjustedTrailingAmount = orderProto.adjustedTrailingAmount
    if orderProto.HasField('adjustableTrailingUnit'): order.adjustableTrailingUnit = orderProto.adjustableTrailingUnit

    softDollarTier = decodeSoftDollarTier(orderProto)
    if softDollarTier is not None: order.softDollarTier = softDollarTier

    if orderProto.HasField('cashQty'): order.cashQty = orderProto.cashQty
    if orderProto.HasField('dontUseAutoPriceForHedge'): order.dontUseAutoPriceForHedge = orderProto.dontUseAutoPriceForHedge
    if orderProto.HasField('isOmsContainer'): order.isOmsContainer = orderProto.isOmsContainer
    if orderProto.HasField('discretionaryUpToLimitPrice'): order.discretionaryUpToLimitPrice = orderProto.discretionaryUpToLimitPrice
    if orderProto.HasField('usePriceMgmtAlgo'): order.usePriceMgmtAlgo = orderProto.usePriceMgmtAlgo
    if orderProto.HasField('duration'): order.duration = orderProto.duration
    if orderProto.HasField('postToAts'): order.postToAts = orderProto.postToAts
    if orderProto.HasField('autoCancelParent'): order.autoCancelParent = orderProto.autoCancelParent
    if orderProto.HasField('minTradeQty'): order.minTradeQty = orderProto.minTradeQty
    if orderProto.HasField('minCompeteSize'): order.minCompeteSize = orderProto.minCompeteSize
    if orderProto.HasField('competeAgainstBestOffset'): order.competeAgainstBestOffset = orderProto.competeAgainstBestOffset
    if orderProto.HasField('midOffsetAtWhole'): order.midOffsetAtWhole = orderProto.midOffsetAtWhole
    if orderProto.HasField('midOffsetAtHalf'): order.midOffsetAtHalf = orderProto.midOffsetAtHalf
    if orderProto.HasField('customerAccount'): order.customerAccount = orderProto.customerAccount
    if orderProto.HasField('professionalCustomer'): order.professionalCustomer = orderProto.professionalCustomer
    if orderProto.HasField('bondAccruedInterest'): order.bondAccruedInterest = orderProto.bondAccruedInterest
    if orderProto.HasField('includeOvernight'): order.includeOvernight = orderProto.includeOvernight
    if orderProto.HasField('extOperator'): order.extOperator = orderProto.extOperator
    if orderProto.HasField('manualOrderIndicator'): order.manualOrderIndicator = orderProto.manualOrderIndicator
    if orderProto.HasField('submitter'): order.submitter = orderProto.submitter
    if orderProto.HasField('imbalanceOnly'): order.imbalanceOnly = orderProto.imbalanceOnly

    return order

def decodeConditions(orderProto: OrderProto) -> list[OrderCondition]:
    orderConditions = []
    orderConditionsProtoList = []
    if orderProto.conditions is not None: orderConditionsProtoList = orderProto.conditions

    if orderConditionsProtoList:
        for orderConditionProto in orderConditionsProtoList:
            conditionType = orderConditionProto.type if orderConditionProto.HasField('type') else 0

            if OrderCondition.Price == conditionType:
                condition = createPriceCondition(orderConditionProto)
            elif OrderCondition.Time == conditionType:
                condition = createTimeCondition(orderConditionProto)
            elif OrderCondition.Margin == conditionType:
                condition = createMarginCondition(orderConditionProto)
            elif OrderCondition.Execution == conditionType:
                condition = createExecutionCondition(orderConditionProto)
            elif OrderCondition.Volume == conditionType:
                condition = createVolumeCondition(orderConditionProto)
            elif OrderCondition.PercentChange == conditionType:
                condition = createPercentChangeCondition(orderConditionProto)

            if condition is not None: orderConditions.append(condition)

    return orderConditions

@staticmethod
def setConditionFields(orderConditionProto: OrderConditionProto, orderCondition: OrderCondition):
    if orderConditionProto.HasField('isConjunctionConnection'): orderCondition.isConjunctionConnection = orderConditionProto.isConjunctionConnection

@staticmethod
def setOperatorConditionFields(orderConditionProto: OrderConditionProto, operatorCondition: OperatorCondition):
    setConditionFields(orderConditionProto, operatorCondition)
    if orderConditionProto.HasField('isMore'): operatorCondition.isMore = orderConditionProto.isMore

@staticmethod
def setContractConditionFields(orderConditionProto: OrderConditionProto, contractCondition: ContractCondition):
    setOperatorConditionFields(orderConditionProto, contractCondition)
    if orderConditionProto.HasField('conId'): contractCondition.conId = orderConditionProto.conId
    if orderConditionProto.HasField('exchange'): contractCondition.exchange = orderConditionProto.exchange

@staticmethod
def createPriceCondition(orderConditionProto: OrderConditionProto) -> PriceCondition:
    priceCondition = PriceCondition()
    setContractConditionFields(orderConditionProto, priceCondition)
    if orderConditionProto.HasField('price'): priceCondition.price = orderConditionProto.price
    if orderConditionProto.HasField('triggerMethod'): priceCondition.triggerMethod = orderConditionProto.triggerMethod
    return priceCondition

@staticmethod
def createTimeCondition(orderConditionProto: OrderConditionProto) -> TimeCondition:
    timeCondition = TimeCondition()
    setOperatorConditionFields(orderConditionProto, timeCondition)
    if orderConditionProto.HasField('time'): timeCondition.time = orderConditionProto.time
    return timeCondition

@staticmethod
def createMarginCondition(orderConditionProto: OrderConditionProto) -> MarginCondition:
    marginCondition = MarginCondition()
    setOperatorConditionFields(orderConditionProto, marginCondition)
    if orderConditionProto.HasField('percent'): marginCondition.percent = orderConditionProto.percent
    return marginCondition

@staticmethod
def createExecutionCondition(orderConditionProto: OrderConditionProto) -> ExecutionCondition:
    executionCondition = ExecutionCondition()
    setConditionFields(orderConditionProto, executionCondition)
    if orderConditionProto.HasField('secType'): executionCondition.secType = orderConditionProto.secType
    if orderConditionProto.HasField('exchange'): executionCondition.exchange = orderConditionProto.exchange
    if orderConditionProto.HasField('symbol'): executionCondition.symbol = orderConditionProto.symbol
    return executionCondition

@staticmethod
def createVolumeCondition(orderConditionProto: OrderConditionProto) -> VolumeCondition:
    volumeCondition = VolumeCondition()
    setContractConditionFields(orderConditionProto, volumeCondition)
    if orderConditionProto.HasField('volume'): volumeCondition.volume = orderConditionProto.volume
    return volumeCondition

@staticmethod
def createPercentChangeCondition(orderConditionProto: OrderConditionProto) -> PercentChangeCondition:
    percentChangeCondition = PercentChangeCondition()
    setContractConditionFields(orderConditionProto, percentChangeCondition)
    if orderConditionProto.HasField('changePercent'): percentChangeCondition.changePercent = orderConditionProto.changePercent
    return percentChangeCondition

@staticmethod
def decodeSoftDollarTier(orderProto: OrderProto) -> SoftDollarTier:
    name = ""
    value = ""
    displayName = ""
    softDollarTierProto = SoftDollarTierProto()
    if orderProto.softDollarTier is not None: softDollarTierProto = orderProto.softDollarTier
    if softDollarTierProto.HasField('name'): name = softDollarTierProto.name
    if softDollarTierProto.HasField('value'): value = softDollarTierProto.value
    if softDollarTierProto.HasField('displayName'): displayName = softDollarTierProto.displayName
    softDollarTier = SoftDollarTier(name, value, displayName)

    return softDollarTier

@staticmethod
def decodeTagValueList(protoMap: dict[str, str]) -> list[TagValue]:
    tagValueList = []
    if protoMap is not None and protoMap:
        for tag, value in protoMap.items():
            tagValue = TagValue()
            tagValue.tag = tag
            tagValue.value = value
            tagValueList.append(tagValue)
    return tagValueList

@staticmethod
def decodeOrderState(orderStateProto: OrderStateProto) -> OrderState:
    orderState = OrderState()
    if orderStateProto.HasField('status'): orderState.status = orderStateProto.status
    if orderStateProto.HasField('initMarginBefore'): orderState.initMarginBefore = orderStateProto.initMarginBefore
    if orderStateProto.HasField('maintMarginBefore'): orderState.maintMarginBefore = decimalMaxString(orderStateProto.maintMarginBefore)
    if orderStateProto.HasField('equityWithLoanBefore'): orderState.equityWithLoanBefore = decimalMaxString(orderStateProto.equityWithLoanBefore)
    if orderStateProto.HasField('initMarginChange'): orderState.initMarginChange = decimalMaxString(orderStateProto.initMarginChange)
    if orderStateProto.HasField('maintMarginChange'): orderState.maintMarginChange = decimalMaxString(orderStateProto.maintMarginChange)
    if orderStateProto.HasField('equityWithLoanChange'): orderState.equityWithLoanChange = decimalMaxString(orderStateProto.equityWithLoanChange)
    if orderStateProto.HasField('initMarginAfter'): orderState.initMarginAfter = decimalMaxString(orderStateProto.initMarginAfter)
    if orderStateProto.HasField('maintMarginAfter'): orderState.maintMarginAfter = decimalMaxString(orderStateProto.maintMarginAfter)
    if orderStateProto.HasField('equityWithLoanAfter'): orderState.equityWithLoanAfter = decimalMaxString(orderStateProto.equityWithLoanAfter)
    if orderStateProto.HasField('commissionAndFees'): orderState.commissionAndFees = orderStateProto.commissionAndFees
    if orderStateProto.HasField('minCommissionAndFees'): orderState.minCommissionAndFees = orderStateProto.minCommissionAndFees
    if orderStateProto.HasField('maxCommissionAndFees'): orderState.maxCommissionAndFees = orderStateProto.maxCommissionAndFees
    if orderStateProto.HasField('commissionAndFeesCurrency'): orderState.commissionAndFeesCurrency = orderStateProto.commissionAndFeesCurrency
    if orderStateProto.HasField('warningText'): orderState.warningText = orderStateProto.warningText
    if orderStateProto.HasField('marginCurrency'): orderState.marginCurrency = orderStateProto.marginCurrency
    if orderStateProto.HasField('initMarginBeforeOutsideRTH'): orderState.initMarginBeforeOutsideRTH = orderStateProto.initMarginBeforeOutsideRTH
    if orderStateProto.HasField('maintMarginBeforeOutsideRTH'): orderState.maintMarginBeforeOutsideRTH = orderStateProto.maintMarginBeforeOutsideRTH
    if orderStateProto.HasField('equityWithLoanBeforeOutsideRTH'): orderState.equityWithLoanBeforeOutsideRTH = orderStateProto.equityWithLoanBeforeOutsideRTH
    if orderStateProto.HasField('initMarginChangeOutsideRTH'): orderState.initMarginChangeOutsideRTH = orderStateProto.initMarginChangeOutsideRTH
    if orderStateProto.HasField('maintMarginChangeOutsideRTH'): orderState.maintMarginChangeOutsideRTH = orderStateProto.maintMarginChangeOutsideRTH
    if orderStateProto.HasField('equityWithLoanChangeOutsideRTH'): orderState.equityWithLoanChangeOutsideRTH = orderStateProto.equityWithLoanChangeOutsideRTH
    if orderStateProto.HasField('initMarginAfterOutsideRTH'): orderState.initMarginAfterOutsideRTH = orderStateProto.initMarginAfterOutsideRTH
    if orderStateProto.HasField('maintMarginAfterOutsideRTH'): orderState.maintMarginAfterOutsideRTH = orderStateProto.maintMarginAfterOutsideRTH
    if orderStateProto.HasField('equityWithLoanAfterOutsideRTH'): orderState.equityWithLoanAfterOutsideRTH = orderStateProto.equityWithLoanAfterOutsideRTH
    if orderStateProto.HasField('suggestedSize'): orderState.suggestedSize = Decimal(orderStateProto.suggestedSize)
    if orderStateProto.HasField('rejectReason'): orderState.rejectReason = orderStateProto.rejectReason

    orderAllocations = decodeOrderAllocations(orderStateProto)
    if orderAllocations is not None and orderAllocations: orderState.orderAllocations = orderAllocations

    return orderState

@staticmethod
def decodeOrderAllocations(orderStateProto: OrderStateProto) -> list[OrderAllocation]:
    orderAllocations = []
    orderAllocationProtoList = []
    if orderStateProto.orderAllocations is not None: orderAllocationProtoList = orderStateProto.orderAllocations
    if orderAllocationProtoList:
        for orderAllocationProto in orderAllocationProtoList:
            orderAllocation = OrderAllocation()
            if orderAllocationProto.HasField('account'): orderAllocation.account = orderAllocationProto.account
            if orderAllocationProto.HasField('position'): orderAllocation.position = Decimal(orderAllocationProto.position)
            if orderAllocationProto.HasField('positionDesired'): orderAllocation.positionDesired = Decimal(orderAllocationProto.positionDesired)
            if orderAllocationProto.HasField('positionAfter'): orderAllocation.positionAfter = Decimal(orderAllocationProto.positionAfter)
            if orderAllocationProto.HasField('desiredAllocQty'): orderAllocation.desiredAllocQty = Decimal(orderAllocationProto.desiredAllocQty)
            if orderAllocationProto.HasField('allowedAllocQty'): orderAllocation.allowedAllocQty = Decimal(orderAllocationProto.allowedAllocQty)
            if orderAllocationProto.HasField('isMonetary'): orderAllocation.isMonetary = orderAllocationProto.isMonetary
            orderAllocations.append(orderAllocation)
    return orderAllocations



import math
import sys
from _decimal import Decimal

NO_VALID_ID = -1
MAX_MSG_LEN = 0xFFFFFF  # 16Mb - 1byte
UNSET_INTEGER = 2**31 - 1
UNSET_DOUBLE = float(sys.float_info.max)
UNSET_LONG = 2**63 - 1
UNSET_DECIMAL = Decimal(2**127 - 1)
DOUBLE_INFINITY = math.inf
INFINITY_STR = "Infinity"



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: GlobalCancelRequest.proto
# Protobuf Python Version: 5.29.3
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    29,
    3,
    '',
    'GlobalCancelRequest.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


import ibapi.protobuf.OrderCancel_pb2 as OrderCancel__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x19GlobalCancelRequest.proto\x12\x08protobuf\x1a\x11OrderCancel.proto\"V\n\x13GlobalCancelRequest\x12/\n\x0borderCancel\x18\x01 \x01(\x0b\x32\x15.protobuf.OrderCancelH\x00\x88\x01\x01\x42\x0e\n\x0c_orderCancelBC\n\x16\x63om.ib.client.protobufB\x18GlobalCancelRequestProto\xaa\x02\x0eIBApi.protobufb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'GlobalCancelRequest_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\026com.ib.client.protobufB\030GlobalCancelRequestProto\252\002\016IBApi.protobuf'
  _globals['_GLOBALCANCELREQUEST']._serialized_start=58
  _globals['_GLOBALCANCELREQUEST']._serialized_end=144
# @@protoc_insertion_point(module_scope)



"""
Copyright (C) 2025 Interactive Brokers LLC. All rights reserved. This code is subject to the terms
 and conditions of the IB API Non-Commercial License or the IB API Commercial License, as applicable.
"""

from ibapi.object_implem import Object
from ibapi.const import UNSET_DECIMAL, UNSET_DOUBLE

from ibapi.utils import intMaxString
from ibapi.utils import floatMaxString
from ibapi.utils import decimalMaxString
from enum import Enum

"""
SAME_POS    = open/close leg value is same as combo
OPEN_POS    = open
CLOSE_POS   = close
UNKNOWN_POS = unknown
"""
(SAME_POS, OPEN_POS, CLOSE_POS, UNKNOWN_POS) = range(4)


class ComboLeg(Object):
    def __init__(self):
        self.conId = 0  # type: int
        self.ratio = 0  # type: int
        self.action = ""  # BUY/SELL/SHORT
        self.exchange = ""
        self.openClose = 0  # type: int # LegOpenClose enum values
        # for stock legs when doing short sale
        self.shortSaleSlot = 0
        self.designatedLocation = ""
        self.exemptCode = -1

    def __str__(self):
        return ",".join(
            (
                intMaxString(self.conId),
                intMaxString(self.ratio),
                str(self.action),
                str(self.exchange),
                intMaxString(self.openClose),
                intMaxString(self.shortSaleSlot),
                str(self.designatedLocation),
                intMaxString(self.exemptCode),
            )
        )


class DeltaNeutralContract(Object):
    def __init__(self):
        self.conId = 0  # type: int
        self.delta = 0.0  # type: float
        self.price = 0.0  # type: float

    def __str__(self):
        return ",".join(
            (str(self.conId), floatMaxString(self.delta), floatMaxString(self.price))
        )


class Contract(Object):
    def __init__(self):
        self.conId = 0
        self.symbol = ""
        self.secType = ""
        self.lastTradeDateOrContractMonth = ""
        self.lastTradeDate = ""
        self.strike = UNSET_DOUBLE  # float !!
        self.right = ""
        self.multiplier = ""
        self.exchange = ""
        self.primaryExchange = ""  # pick an actual (ie non-aggregate) exchange that the contract trades on.
        # DO NOT SET TO SMART.
        self.currency = ""
        self.localSymbol = ""
        self.tradingClass = ""
        self.includeExpired = False
        self.secIdType = ""  # CUSIP;SEDOL;ISIN;RIC
        self.secId = ""
        self.description = ""
        self.issuerId = ""

        # combos
        self.comboLegsDescrip = (
            ""
        )  # type: str #received in open order 14 and up for all combos
        self.comboLegs = []  # type: list[ComboLeg]
        self.deltaNeutralContract = None

    def __str__(self):
        s = (
            "ConId: %s, Symbol: %s, SecType: %s, LastTradeDateOrContractMonth: %s, Strike: %s, Right: %s, Multiplier: %s, Exchange: %s, PrimaryExchange: %s, "
            "Currency: %s, LocalSymbol: %s, TradingClass: %s, IncludeExpired: %s, SecIdType: %s, SecId: %s, Description: %s, "
            "IssuerId: %s"
            % (
                intMaxString(self.conId),
                str(self.symbol),
                str(self.secType),
                str(self.lastTradeDateOrContractMonth),
                floatMaxString(self.strike),
                str(self.right),
                str(self.multiplier),
                str(self.exchange),
                str(self.primaryExchange),
                str(self.currency),
                str(self.localSymbol),
                str(self.tradingClass),
                str(self.includeExpired),
                str(self.secIdType),
                str(self.secId),
                str(self.description),
                str(self.issuerId),
            )
        )
        s += "Combo:" + self.comboLegsDescrip

        if self.comboLegs:
            for leg in self.comboLegs:
                s += ";" + str(leg)

        if self.deltaNeutralContract:
            s += ";" + str(self.deltaNeutralContract)

        return s


class ContractDetails(Object):
    def __init__(self):
        self.contract = Contract()
        self.marketName = ""
        self.minTick = 0.0
        self.orderTypes = ""
        self.validExchanges = ""
        self.priceMagnifier = 0
        self.underConId = 0
        self.longName = ""
        self.contractMonth = ""
        self.industry = ""
        self.category = ""
        self.subcategory = ""
        self.timeZoneId = ""
        self.tradingHours = ""
        self.liquidHours = ""
        self.evRule = ""
        self.evMultiplier = 0
        self.aggGroup = 0
        self.underSymbol = ""
        self.underSecType = ""
        self.marketRuleIds = ""
        self.secIdList = None
        self.realExpirationDate = ""
        self.lastTradeTime = ""
        self.stockType = ""
        self.minSize = UNSET_DECIMAL
        self.sizeIncrement = UNSET_DECIMAL
        self.suggestedSizeIncrement = UNSET_DECIMAL
        # BOND values
        self.cusip = ""
        self.ratings = ""
        self.descAppend = ""
        self.bondType = ""
        self.couponType = ""
        self.callable = False
        self.putable = False
        self.coupon = 0
        self.convertible = False
        self.maturity = ""
        self.issueDate = ""
        self.nextOptionDate = ""
        self.nextOptionType = ""
        self.nextOptionPartial = False
        self.notes = ""
        # FUND values
        self.fundName = ""
        self.fundFamily = ""
        self.fundType = ""
        self.fundFrontLoad = ""
        self.fundBackLoad = ""
        self.fundBackLoadTimeInterval = ""
        self.fundManagementFee = ""
        self.fundClosed = False
        self.fundClosedForNewInvestors = False
        self.fundClosedForNewMoney = False
        self.fundNotifyAmount = ""
        self.fundMinimumInitialPurchase = ""
        self.fundSubsequentMinimumPurchase = ""
        self.fundBlueSkyStates = ""
        self.fundBlueSkyTerritories = ""
        self.fundDistributionPolicyIndicator = FundDistributionPolicyIndicator.NoneItem
        self.fundAssetType = FundAssetType.NoneItem
        self.ineligibilityReasonList = None

    def __str__(self):
        s = ",".join(
            (
                str(self.contract),
                str(self.marketName),
                floatMaxString(self.minTick),
                str(self.orderTypes),
                str(self.validExchanges),
                intMaxString(self.priceMagnifier),
                intMaxString(self.underConId),
                str(self.longName),
                str(self.contractMonth),
                str(self.industry),
                str(self.category),
                str(self.subcategory),
                str(self.timeZoneId),
                str(self.tradingHours),
                str(self.liquidHours),
                str(self.evRule),
                intMaxString(self.evMultiplier),
                str(self.underSymbol),
                str(self.underSecType),
                str(self.marketRuleIds),
                intMaxString(self.aggGroup),
                str(self.secIdList),
                str(self.realExpirationDate),
                str(self.stockType),
                str(self.cusip),
                str(self.ratings),
                str(self.descAppend),
                str(self.bondType),
                str(self.couponType),
                str(self.callable),
                str(self.putable),
                str(self.coupon),
                str(self.convertible),
                str(self.maturity),
                str(self.issueDate),
                str(self.nextOptionDate),
                str(self.nextOptionType),
                str(self.nextOptionPartial),
                str(self.notes),
                decimalMaxString(self.minSize),
                decimalMaxString(self.sizeIncrement),
                decimalMaxString(self.suggestedSizeIncrement),
                str(self.ineligibilityReasonList),
            )
        )

        return s


class ContractDescription(Object):
    def __init__(self):
        self.contract = Contract()
        self.derivativeSecTypes = []  # type: list[str]

class FundAssetType(Enum):
    NoneItem = ("None", "None")
    Others = ("000", "Others"), 
    MoneyMarket = ("001", "Money Market")
    FixedIncome = ("002", "Fixed Income")
    MultiAsset = ("003", "Multi-asset")
    Equity = ("004", "Equity")
    Sector = ("005", "Sector")
    Guaranteed = ("006", "Guaranteed")
    Alternative = ("007", "Alternative")

class FundDistributionPolicyIndicator(Enum):
    NoneItem = ("None", "None")
    AccumulationFund = ("N", "Accumulation Fund")
    IncomeFund = ("Y", "Income Fund")



"""
Copyright (C) 2025 Interactive Brokers LLC. All rights reserved. This code is subject to the terms
 and conditions of the IB API Non-Commercial License or the IB API Commercial License, as applicable.
"""

"""
The known server versions.
"""

# MIN_SERVER_VER_REAL_TIME_BARS       = 34
# MIN_SERVER_VER_SCALE_ORDERS         = 35
# MIN_SERVER_VER_SNAPSHOT_MKT_DATA    = 35
# MIN_SERVER_VER_SSHORT_COMBO_LEGS    = 35
# MIN_SERVER_VER_WHAT_IF_ORDERS       = 36
# MIN_SERVER_VER_CONTRACT_CONID       = 37
MIN_SERVER_VER_PTA_ORDERS = 39
MIN_SERVER_VER_FUNDAMENTAL_DATA = 40
MIN_SERVER_VER_DELTA_NEUTRAL = 40
MIN_SERVER_VER_CONTRACT_DATA_CHAIN = 40
MIN_SERVER_VER_SCALE_ORDERS2 = 40
MIN_SERVER_VER_ALGO_ORDERS = 41
MIN_SERVER_VER_EXECUTION_DATA_CHAIN = 42
MIN_SERVER_VER_NOT_HELD = 44
MIN_SERVER_VER_SEC_ID_TYPE = 45
MIN_SERVER_VER_PLACE_ORDER_CONID = 46
MIN_SERVER_VER_REQ_MKT_DATA_CONID = 47
MIN_SERVER_VER_REQ_CALC_IMPLIED_VOLAT = 49
MIN_SERVER_VER_REQ_CALC_OPTION_PRICE = 50
MIN_SERVER_VER_SSHORTX_OLD = 51
MIN_SERVER_VER_SSHORTX = 52
MIN_SERVER_VER_REQ_GLOBAL_CANCEL = 53
MIN_SERVER_VER_HEDGE_ORDERS = 54
MIN_SERVER_VER_REQ_MARKET_DATA_TYPE = 55
MIN_SERVER_VER_OPT_OUT_SMART_ROUTING = 56
MIN_SERVER_VER_SMART_COMBO_ROUTING_PARAMS = 57
MIN_SERVER_VER_DELTA_NEUTRAL_CONID = 58
MIN_SERVER_VER_SCALE_ORDERS3 = 60
MIN_SERVER_VER_ORDER_COMBO_LEGS_PRICE = 61
MIN_SERVER_VER_TRAILING_PERCENT = 62
MIN_SERVER_VER_DELTA_NEUTRAL_OPEN_CLOSE = 66
MIN_SERVER_VER_POSITIONS = 67
MIN_SERVER_VER_ACCOUNT_SUMMARY = 67
MIN_SERVER_VER_TRADING_CLASS = 68
MIN_SERVER_VER_SCALE_TABLE = 69
MIN_SERVER_VER_LINKING = 70
MIN_SERVER_VER_ALGO_ID = 71
MIN_SERVER_VER_OPTIONAL_CAPABILITIES = 72
MIN_SERVER_VER_ORDER_SOLICITED = 73
MIN_SERVER_VER_LINKING_AUTH = 74
MIN_SERVER_VER_PRIMARYEXCH = 75
MIN_SERVER_VER_RANDOMIZE_SIZE_AND_PRICE = 76
MIN_SERVER_VER_FRACTIONAL_POSITIONS = 101
MIN_SERVER_VER_PEGGED_TO_BENCHMARK = 102
MIN_SERVER_VER_MODELS_SUPPORT = 103
MIN_SERVER_VER_SEC_DEF_OPT_PARAMS_REQ = 104
MIN_SERVER_VER_EXT_OPERATOR = 105
MIN_SERVER_VER_SOFT_DOLLAR_TIER = 106
MIN_SERVER_VER_REQ_FAMILY_CODES = 107
MIN_SERVER_VER_REQ_MATCHING_SYMBOLS = 108
MIN_SERVER_VER_PAST_LIMIT = 109
MIN_SERVER_VER_MD_SIZE_MULTIPLIER = 110
MIN_SERVER_VER_CASH_QTY = 111
MIN_SERVER_VER_REQ_MKT_DEPTH_EXCHANGES = 112
MIN_SERVER_VER_TICK_NEWS = 113
MIN_SERVER_VER_REQ_SMART_COMPONENTS = 114
MIN_SERVER_VER_REQ_NEWS_PROVIDERS = 115
MIN_SERVER_VER_REQ_NEWS_ARTICLE = 116
MIN_SERVER_VER_REQ_HISTORICAL_NEWS = 117
MIN_SERVER_VER_REQ_HEAD_TIMESTAMP = 118
MIN_SERVER_VER_REQ_HISTOGRAM = 119
MIN_SERVER_VER_SERVICE_DATA_TYPE = 120
MIN_SERVER_VER_AGG_GROUP = 121
MIN_SERVER_VER_UNDERLYING_INFO = 122
MIN_SERVER_VER_CANCEL_HEADTIMESTAMP = 123
MIN_SERVER_VER_SYNT_REALTIME_BARS = 124
MIN_SERVER_VER_CFD_REROUTE = 125
MIN_SERVER_VER_MARKET_RULES = 126
MIN_SERVER_VER_PNL = 127
MIN_SERVER_VER_NEWS_QUERY_ORIGINS = 128
MIN_SERVER_VER_UNREALIZED_PNL = 129
MIN_SERVER_VER_HISTORICAL_TICKS = 130
MIN_SERVER_VER_MARKET_CAP_PRICE = 131
MIN_SERVER_VER_PRE_OPEN_BID_ASK = 132
MIN_SERVER_VER_REAL_EXPIRATION_DATE = 134
MIN_SERVER_VER_REALIZED_PNL = 135
MIN_SERVER_VER_LAST_LIQUIDITY = 136
MIN_SERVER_VER_TICK_BY_TICK = 137
MIN_SERVER_VER_DECISION_MAKER = 138
MIN_SERVER_VER_MIFID_EXECUTION = 139
MIN_SERVER_VER_TICK_BY_TICK_IGNORE_SIZE = 140
MIN_SERVER_VER_AUTO_PRICE_FOR_HEDGE = 141
MIN_SERVER_VER_WHAT_IF_EXT_FIELDS = 142
MIN_SERVER_VER_SCANNER_GENERIC_OPTS = 143
MIN_SERVER_VER_API_BIND_ORDER = 144
MIN_SERVER_VER_ORDER_CONTAINER = 145
MIN_SERVER_VER_SMART_DEPTH = 146
MIN_SERVER_VER_REMOVE_NULL_ALL_CASTING = 147
MIN_SERVER_VER_D_PEG_ORDERS = 148
MIN_SERVER_VER_MKT_DEPTH_PRIM_EXCHANGE = 149
MIN_SERVER_VER_COMPLETED_ORDERS = 150
MIN_SERVER_VER_PRICE_MGMT_ALGO = 151
MIN_SERVER_VER_STOCK_TYPE = 152
MIN_SERVER_VER_ENCODE_MSG_ASCII7 = 153
MIN_SERVER_VER_SEND_ALL_FAMILY_CODES = 154
MIN_SERVER_VER_NO_DEFAULT_OPEN_CLOSE = 155
MIN_SERVER_VER_PRICE_BASED_VOLATILITY = 156
MIN_SERVER_VER_REPLACE_FA_END = 157
MIN_SERVER_VER_DURATION = 158
MIN_SERVER_VER_MARKET_DATA_IN_SHARES = 159
MIN_SERVER_VER_POST_TO_ATS = 160
MIN_SERVER_VER_WSHE_CALENDAR = 161
MIN_SERVER_VER_AUTO_CANCEL_PARENT = 162
MIN_SERVER_VER_FRACTIONAL_SIZE_SUPPORT = 163
MIN_SERVER_VER_SIZE_RULES = 164
MIN_SERVER_VER_HISTORICAL_SCHEDULE = 165
MIN_SERVER_VER_ADVANCED_ORDER_REJECT = 166
MIN_SERVER_VER_USER_INFO = 167
MIN_SERVER_VER_CRYPTO_AGGREGATED_TRADES = 168
MIN_SERVER_VER_MANUAL_ORDER_TIME = 169
MIN_SERVER_VER_PEGBEST_PEGMID_OFFSETS = 170
MIN_SERVER_VER_WSH_EVENT_DATA_FILTERS = 171
MIN_SERVER_VER_IPO_PRICES = 172
MIN_SERVER_VER_WSH_EVENT_DATA_FILTERS_DATE = 173
MIN_SERVER_VER_INSTRUMENT_TIMEZONE = 174
MIN_SERVER_VER_HMDS_MARKET_DATA_IN_SHARES = 175
MIN_SERVER_VER_BOND_ISSUERID = 176
MIN_SERVER_VER_FA_PROFILE_DESUPPORT = 177
MIN_SERVER_VER_PENDING_PRICE_REVISION = 178
MIN_SERVER_VER_FUND_DATA_FIELDS = 179
MIN_SERVER_VER_MANUAL_ORDER_TIME_EXERCISE_OPTIONS = 180
MIN_SERVER_VER_OPEN_ORDER_AD_STRATEGY = 181
MIN_SERVER_VER_LAST_TRADE_DATE = 182
MIN_SERVER_VER_CUSTOMER_ACCOUNT = 183
MIN_SERVER_VER_PROFESSIONAL_CUSTOMER = 184
MIN_SERVER_VER_BOND_ACCRUED_INTEREST = 185
MIN_SERVER_VER_INELIGIBILITY_REASONS = 186
MIN_SERVER_VER_RFQ_FIELDS = 187
MIN_SERVER_VER_BOND_TRADING_HOURS = 188
MIN_SERVER_VER_INCLUDE_OVERNIGHT = 189
MIN_SERVER_VER_UNDO_RFQ_FIELDS = 190
MIN_SERVER_VER_PERM_ID_AS_LONG = 191
MIN_SERVER_VER_CME_TAGGING_FIELDS = 192
MIN_SERVER_VER_CME_TAGGING_FIELDS_IN_OPEN_ORDER = 193
MIN_SERVER_VER_ERROR_TIME = 194
MIN_SERVER_VER_FULL_ORDER_PREVIEW_FIELDS = 195
MIN_SERVER_VER_HISTORICAL_DATA_END = 196
MIN_SERVER_VER_CURRENT_TIME_IN_MILLIS = 197
MIN_SERVER_VER_SUBMITTER = 198
MIN_SERVER_VER_IMBALANCE_ONLY = 199
MIN_SERVER_VER_PARAMETRIZED_DAYS_OF_EXECUTIONS = 200
MIN_SERVER_VER_PROTOBUF = 201
MIN_SERVER_VER_ZERO_STRIKE = 202
MIN_SERVER_VER_PROTOBUF_PLACE_ORDER = 203

# 100+ messaging */
# 100 = enhanced handshake, msg length prefixes

MIN_CLIENT_VER = 100
MAX_CLIENT_VER = MIN_SERVER_VER_PROTOBUF_PLACE_ORDER
